import os

import supervisely.convert.image.pdf.pdf_helper as helper
from supervisely import Annotation, ProjectMeta, logger
from supervisely.convert.base_converter import AvailableImageConverters
from supervisely.convert.image.image_converter import ImageConverter
from supervisely.io.fs import (
    JUNK_FILES,
    get_file_ext,
    get_file_name,
    mkdir,
    silent_remove,
)
from supervisely.io.json import load_json_file


class PDFConverter(ImageConverter):

    def __str__(self):
        return AvailableImageConverters.PDF

    def validate_format(self) -> bool:
        detected_pdf_cnt = 0
        images_list = []
        for root, _, files in os.walk(self._input_data):
            for file in files:
                full_path = os.path.join(root, file)
                ext = get_file_ext(full_path)
                if file in JUNK_FILES:
                    continue
                elif ext == ".pdf":
                    images_list.append(full_path)
                    detected_pdf_cnt += 1
                else:
                    continue

        if detected_pdf_cnt == 0:
            return False
        else:
            # create Items
            self._items = []
            for image_path in images_list:
                dir_path, pdf_name = os.path.split(image_path)
                pdf_name_without_ext = get_file_name(pdf_name)
                pdf_dir_path = os.path.join(dir_path, pdf_name_without_ext)
                mkdir(pdf_dir_path)

                # @TODO: add progress for pdf with many pages?
                success = helper.pages_to_images(
                    doc_path=image_path,
                    save_path=pdf_dir_path,
                    dpi=300,
                    logger=logger,
                )
                silent_remove(image_path)

                if not success:
                    continue

                for root, _, files in os.walk(pdf_dir_path):
                    for file in files:
                        image_path = os.path.join(root, file)
                        item = self.Item(image_path)
                        self._items.append(item)
            return True

    def to_supervisely(
        self,
        item: ImageConverter.Item,
        meta: ProjectMeta = None,
        renamed_classes: dict = None,
        renamed_tags: dict = None,
    ) -> Annotation:
        """Convert to Supervisely format."""
        return item.create_empty_annotation()



# coding: utf-8

import os
import struct
import requests
import traceback
import time
from logging import Logger

from supervisely.worker_api.retriers import retriers_from_cfg
from supervisely.io.network_exceptions import process_requests_exception, process_unhandled_request


class AgentAPI:
    def __init__(self, token, server_address: str, ext_logger: Logger, cfg_path=None):
        self.logger = ext_logger
        self._base_server_adress = server_address
        if ("http://" not in self._base_server_adress) and (
            "https://" not in self._base_server_adress
        ):
            self._base_server_adress = os.path.join("http://", self._base_server_adress)
        self.server_address = os.path.join(self._base_server_adress, "agent")
        self.headers = {
            "Content-type": "application/octet-stream",
            "Accept-Encoding": "deflate",  # to override default 'Accept-Encoding': 'gzip, deflate'
        }
        if token is not None:
            self.headers["x-token"] = token
        self._retriers = retriers_from_cfg(cfg_path)

        self._require_https_redirect_check = not self._base_server_adress.startswith("https://")

    def _get_retrier(self, api_method_name, common_selector):
        retrier = self._retriers.get(api_method_name)
        if retrier is None:
            retrier = self._retriers[common_selector]
        return retrier

    def add_to_metadata(self, key, value):
        self.headers[key] = value

    def rm_from_metadata(self, key):
        self.headers.pop(key, None)

    def _send_request(self, api_method_name, request_data, timeout, in_stream, addit_headers):
        self._check_https_redirect()
        url = os.path.join(self.server_address, api_method_name)
        if not addit_headers:
            addit_headers = {}
        cur_header = {**self.headers, **addit_headers}
        not_log_request = api_method_name != "Log"
        server_reply = None
        try:
            server_reply = requests.post(
                url, headers=cur_header, data=request_data, stream=in_stream, timeout=timeout
            )
            server_reply.raise_for_status()
        except requests.RequestException as exc:
            process_requests_exception(
                self.logger,
                exc,
                api_method_name,
                url,
                verbose=not_log_request,
                swallow_exc=False,
                response=server_reply,
            )
        except Exception as exc:
            process_unhandled_request(self.logger, exc)

        return server_reply

    # magic value 4 means four bytes for message length
    # http://www.sureshjoshi.com/development/streaming-protocol-buffers/
    # https://www.datadoghq.com/blog/engineering/protobuf-parsing-in-python/
    def _get_input_stream(
        self, api_method_name, res_proto_fn, request_data, timeout, addit_headers
    ):
        def cut_len(msg_buff_):
            cur_m_len = struct.unpack(">I", msg_buff_[0:4])[0]
            return cur_m_len, msg_buff_[4:]

        def append_to_msg_buffer(msg_len_, msg_buf_, rest_buf):
            if msg_len_ > len(msg_buf_) + len(rest_buf):
                msg_buf_ = msg_buf_ + rest_buf
                rest_buf = b""
            else:
                tmp_cut_len = msg_len_ - len(msg_buf_)
                msg_buf_ = msg_buf_ + rest_buf[0:tmp_cut_len]
                rest_buf = rest_buf[tmp_cut_len:]
            return msg_buf_, rest_buf

        # request package crashes on empty chunks
        try:
            with self._send_request(
                api_method_name, request_data, timeout, in_stream=True, addit_headers=addit_headers
            ) as reply:
                msg_len = None
                msg_buf = b""
                for buffer in reply.iter_content(chunk_size=None):
                    while len(buffer) > 0:
                        if msg_len is None:
                            if len(msg_buf) < 4:
                                msg_buf, buffer = append_to_msg_buffer(4, msg_buf, buffer)
                            if len(msg_buf) < 4:
                                continue
                            msg_len, msg_buf = cut_len(msg_buf)
                        msg_buf, buffer = append_to_msg_buffer(msg_len, msg_buf, buffer)
                        if msg_len == len(msg_buf):
                            if msg_len == 0:
                                pass
                            else:
                                proto_msg = res_proto_fn()
                                proto_msg.ParseFromString(msg_buf)
                                yield proto_msg
                            msg_len = None
                            msg_buf = b""
                if msg_len is not None:
                    raise RuntimeError("MISSED_STREAM_CHUNKS")
        except requests.exceptions.ChunkedEncodingError:
            raise RuntimeError("Unknown error during stream. Please contact support.")
        except requests.RequestException:
            raise
        except Exception as e:
            self.logger.error(traceback.format_exc(), exc_info=True, extra={"exc_str": str(e)})
            raise e

    def _put_out_stream(
        self, api_method_name, res_proto_fn, chunk_generator, timeout, addit_headers
    ):
        def bindata_generator():
            for chunk in chunk_generator:
                size = chunk.ByteSize()
                res_bytes_with_len = struct.pack(">I", size) + chunk.SerializeToString()
                yield res_bytes_with_len

        resp = self._send_request(
            api_method_name,
            bindata_generator(),
            timeout,
            in_stream=False,
            addit_headers=addit_headers,
        )
        res_proto = res_proto_fn()
        res_proto.ParseFromString(resp.content)
        return res_proto

    # will not log it now
    def simple_request(self, api_method_name, res_proto_fn, proto_request, addit_headers=None):
        data_to_send = proto_request.SerializeToString()
        retrier = self._get_retrier(api_method_name, "__simple_request")
        resp = retrier.request(
            self._send_request,
            api_method_name,
            data_to_send,
            in_stream=False,
            addit_headers=addit_headers,
        )
        if resp is None:
            return None  # swallowed exception
        res_proto = res_proto_fn()
        res_proto.ParseFromString(resp.content)
        return res_proto

    def get_stream_with_data(
        self, api_method_name, res_proto_fn, proto_request, addit_headers=None
    ):
        data_to_send = proto_request.SerializeToString()
        retrier = self._get_retrier(api_method_name, "__data_stream_in")
        yield from retrier.request(
            self._get_input_stream,
            api_method_name,
            res_proto_fn,
            data_to_send,
            addit_headers=addit_headers,
        )

    def get_endless_stream(
        self,
        api_method_name,
        res_proto_fn,
        proto_request,
        addit_headers=None,
        server_fail_limit=10,
        wait_server_sec=10,
    ):
        data_to_send = proto_request.SerializeToString()
        for attempt in range(server_fail_limit):
            retrier = self._get_retrier(api_method_name, "__endless_stream_in")
            yield from retrier.request(
                self._get_input_stream,
                api_method_name,
                res_proto_fn,
                data_to_send,
                addit_headers=addit_headers,
            )
            self.logger.warn("Endless input stream end", extra={"method": api_method_name})
            if attempt != server_fail_limit - 1:
                time.sleep(wait_server_sec)

    def put_stream_with_data(
        self, api_method_name, res_proto_fn, chunk_generator, addit_headers=None
    ):
        retrier = self._get_retrier(api_method_name, "__data_stream_out")
        res = retrier.request(
            self._put_out_stream,
            api_method_name,
            res_proto_fn,
            chunk_generator,
            addit_headers=addit_headers,
        )
        return res

    @classmethod
    def catch_http_err(cls, code_list, fn, *args, **kwargs):
        try:
            res = fn(*args, **kwargs)
            return res, None
        except requests.exceptions.HTTPError as e:
            res_code = e.response.status_code
            if res_code in code_list:
                return None, res_code
            raise

    def _check_https_redirect(self):
        if self._require_https_redirect_check is True:
            response = requests.get(self._base_server_adress, allow_redirects=False)
            if (300 <= response.status_code < 400) or (
                response.headers.get("Location", "").startswith("https://")
            ):
                self._base_server_adress = self._base_server_adress.replace("http://", "https://")
                self.server_address = self.server_address.replace("http://", "https://")
                msg = (
                    "You're using HTTP server address on agent while the server requires HTTPS. "
                    "Supervisely automatically changed the server address to HTTPS for you. "
                    f"Consider updating your server address to {self._base_server_adress}"
                )
                self.logger.warn(msg)

            self._require_https_redirect_check = False



import json
import os
import shutil
from enum import Enum
from logging import Logger
from pathlib import Path
from threading import Lock, Thread
from time import sleep
from typing import Any, Callable, Generator, List, Optional, Tuple, Union

import cv2
import numpy as np
from cacheout import Cache as CacheOut
from cachetools import Cache, LRUCache, TTLCache
from fastapi import BackgroundTasks, FastAPI, Form, Request, UploadFile

import supervisely as sly
from supervisely.io.fs import silent_remove


class PersistentImageLRUCache(LRUCache):
    __marker = object()

    def __init__(self, maxsize, filepath: Path, getsizeof=None):
        super().__init__(maxsize)
        self._base_dir = filepath

    def __getitem__(self, key: Any) -> Any:
        filepath = super(PersistentImageLRUCache, self).__getitem__(key)
        return sly.image.read(str(filepath))

    def __setitem__(self, key: Any, value: Any) -> None:
        if not self._base_dir.exists():
            self._base_dir.mkdir()

        filepath = self._base_dir / f"{str(key)}.png"
        super(PersistentImageLRUCache, self).__setitem__(key, filepath)

        if filepath.exists():
            sly.logger.debug(f"Rewrite image {str(filepath)}")
        sly.image.write(str(filepath), value)

    def pop(self, key, default=__marker):
        if key in self:
            filepath = self._base_dir / f"{str(key)}.png"
            value = self[key]
            del self[key]
            silent_remove(filepath)
            sly.logger.debug(f"Remove {filepath} frame")
        elif default is self.__marker:
            raise KeyError(key)
        else:
            value = default
        return value

    def clear(self, rm_base_folder=True) -> None:
        while self.currsize > 0:
            self.popitem()
        if rm_base_folder:
            shutil.rmtree(self._base_dir)


class PersistentImageTTLCache(TTLCache):
    def __init__(self, maxsize: int, ttl: int, filepath: Path):
        super().__init__(maxsize, ttl)
        self._base_dir = filepath

    def __delitem__(self, key: Any) -> None:
        self.__del_file(key)
        return super().__delitem__(key)

    def __del_file(self, key: str):
        cache_getitem = Cache.__getitem__
        filepath = cache_getitem(self, key)
        try:
            silent_remove(filepath)
        except TypeError:
            pass

    def __get_keys(self):
        # pylint: disable=no-member
        return self._TTLCache__links.keys()

    def expire(self, time=None):
        """Remove expired items from the cache."""
        existing = set(self.__get_keys())
        if time is None:
            # pylint: disable=no-member
            time = self._TTLCache__timer()
        # pylint: disable=no-member
        root = self._TTLCache__root
        curr = root.next
        # pylint: disable=no-member
        links = self._TTLCache__links
        cache_delitem = Cache.__delitem__
        while curr is not root and curr.expire < time:
            self.__del_file(curr.key)
            cache_delitem(self, curr.key)
            del links[curr.key]
            next = curr.next
            curr.unlink()
            curr = next
        deleted = existing.difference(self.__get_keys())
        if len(deleted) > 0:
            sly.logger.debug(f"Deleted keys: {deleted}")

    def clear(self, rm_base_folder=True) -> None:
        while self.currsize > 0:
            self.popitem()
        if rm_base_folder:
            shutil.rmtree(self._base_dir)

    def save_image(self, key, image: np.ndarray) -> None:
        if not self._base_dir.exists():
            self._base_dir.mkdir()

        filepath = self._base_dir / f"{str(key)}.png"
        super(PersistentImageTTLCache, self).__setitem__(key, filepath)

        if filepath.exists():
            sly.logger.debug(f"Rewrite image {str(filepath)}")
        sly.image.write(str(filepath), image)

    def get_image_path(self, key: Any) -> Path:
        return super().__getitem__(key)

    def get_image(self, key: Any):
        return sly.image.read(str(self[key]))

    def save_video(self, video_id: int, src_video_path: str) -> None:
        video_path = self._base_dir / f"video_{video_id}.{src_video_path.split('.')[-1]}"
        if src_video_path != str(video_path):
            shutil.move(src_video_path, str(video_path))
        super().__setitem__(video_id, video_path)

    def get_video_path(self, video_id: int) -> Path:
        return super().__getitem__(video_id)

    def save_project_meta(self, key, value):
        super().__setitem__(key, value)

    def get_project_meta(self, project_meta_name):
        return super().__getitem__(project_meta_name)


class InferenceImageCache:
    class _LoadType(Enum):
        ImageId: str = "IMAGE"
        ImageHash: str = "HASH"
        Frame: str = "FRAME"
        Video: str = "VIDEO"

    def __init__(
        self,
        maxsize: int,
        ttl: int,
        is_persistent: bool = True,
        base_folder: str = sly.env.smart_cache_container_dir(),
    ) -> None:
        self.is_persistent = is_persistent
        self._maxsize = maxsize
        self._ttl = ttl
        self._lock = Lock()
        self._load_queue = CacheOut(10 * 60)

        if is_persistent:
            self._data_dir = Path(base_folder)
            self._data_dir.mkdir(parents=True, exist_ok=True)
            self._cache = PersistentImageTTLCache(maxsize, ttl, self._data_dir)
        else:
            self._cache = TTLCache(maxsize, ttl)

    def clear_cache(self):
        with self._lock:
            self._cache.clear(False)

    def download_image(self, api: sly.Api, image_id: int):
        name = self._image_name(image_id)
        self._wait_if_in_queue(name, api.logger)

        if name not in self._cache:
            self._load_queue.set(name, image_id)
            api.logger.debug(f"Add image #{image_id} to cache")
            img = api.image.download_np(image_id)
            self._add_to_cache(name, img)
            return img

        api.logger.debug(f"Get image #{image_id} from cache")
        return self._cache.get_image(name)

    def download_images(self, api: sly.Api, dataset_id: int, image_ids: List[int], **kwargs):
        return_images = kwargs.get("return_images", True)

        def load_generator(image_ids: List[int]):
            return api.image.download_nps_generator(dataset_id, image_ids)

        return self._download_many(
            image_ids,
            self._image_name,
            load_generator,
            api.logger,
            return_images,
        )

    def download_image_by_hash(self, api: sly.Api, img_hash: str) -> np.ndarray:
        image_key = self._image_name(img_hash)
        self._wait_if_in_queue(image_key, api.logger)

        if image_key not in self._cache:
            self._load_queue.set(image_key, img_hash)
            image = api.image.download_nps_by_hashes([img_hash])
            self._add_to_cache(image_key, image)
            return image
        return self._cache.get_image(image_key)

    def download_images_by_hashes(
        self, api: sly.Api, img_hashes: List[str], **kwargs
    ) -> List[np.ndarray]:
        return_images = kwargs.get("return_images", True)

        def load_generator(img_hashes: List[str]):
            return api.image.download_nps_by_hashes_generator(img_hashes)

        return self._download_many(
            img_hashes,
            self._image_name,
            load_generator,
            api.logger,
            return_images,
        )

    def _read_frames_from_cached_video(
        self, video_id: int, frame_indexes: List[int]
    ) -> List[np.ndarray]:
        video_path = self._cache.get_video_path(video_id)
        if video_path is None or not video_path.exists():
            raise KeyError(f"Video {video_id} not found in cache")
        cap = cv2.VideoCapture(str(video_path))
        frames = []
        prev_idx = -1
        for frame_index in frame_indexes:
            if frame_index != prev_idx + 1:
                cap.set(cv2.CAP_PROP_POS_FRAMES, frame_index)
            ret, frame = cap.read()
            if not ret:
                raise KeyError(f"Frame {frame_index} not found in video {video_id}")
            frames.append(cv2.cvtColor(frame, cv2.COLOR_BGR2RGB))
            prev_idx = frame_index
        cap.release()
        return frames

    def get_frame_from_cache(self, video_id: int, frame_index: int) -> np.ndarray:
        name = self._frame_name(video_id, frame_index)
        if isinstance(self._cache, PersistentImageTTLCache):
            if name in self._cache:
                return self._cache.get_image(name)
            return self._read_frames_from_cached_video(video_id, [frame_index])[0]
        frame = self._cache.get(name)
        if frame is None:
            raise KeyError(f"Frame {frame_index} not found in video {video_id}")

    def get_frames_from_cache(self, video_id: int, frame_indexes: List[int]) -> List[np.ndarray]:
        if isinstance(self._cache, PersistentImageTTLCache) and video_id in self._cache:
            return self._read_frames_from_cached_video(video_id, frame_indexes)
        else:
            return [
                self.get_frame_from_cache(video_id, frame_index) for frame_index in frame_indexes
            ]

    def download_frame(self, api: sly.Api, video_id: int, frame_index: int) -> np.ndarray:
        name = self._frame_name(video_id, frame_index)
        self._wait_if_in_queue(name, api.logger)

        if name not in self._cache:
            if video_id in self._cache:
                api.logger.debug(
                    f"Get frame #{frame_index} for video #{video_id} from cache (video file)"
                )
                try:
                    return self.get_frame_from_cache(video_id, frame_index)
                except:
                    sly.logger.warning(
                        f"Frame {frame_index} not found in video {video_id}", exc_info=True
                    )

            self._load_queue.set(name, (video_id, frame_index))
            frame = api.video.frame.download_np(video_id, frame_index)
            self._add_to_cache(name, frame)
            api.logger.debug(f"Add frame #{frame_index} for video #{video_id} to cache")
            return frame

        api.logger.debug(f"Get frame #{frame_index} for video #{video_id} from cache")
        return self._cache.get_image(name)

    def download_frames(
        self, api: sly.Api, video_id: int, frame_indexes: List[int], **kwargs
    ) -> List[np.ndarray]:
        return_images = kwargs.get("return_images", True)

        if video_id in self._cache:
            try:
                return self.get_frames_from_cache(video_id, frame_indexes)
            except:
                sly.logger.warning(
                    f"Frames {frame_indexes} not found in video {video_id}", exc_info=True
                )
                Thread(target=self.download_video, args=(api, video_id)).start()

        def name_constuctor(frame_index: int):
            return self._frame_name(video_id, frame_index)

        def load_generator(frame_indexes: List[int]):
            return api.video.frame.download_nps_generator(video_id, frame_indexes)

        return self._download_many(
            frame_indexes,
            name_constuctor,
            load_generator,
            api.logger,
            return_images,
        )

    def add_video_to_cache(self, video_id: int, video_path: Path) -> None:
        """
        Adds video to cache.
        """
        if isinstance(self._cache, PersistentImageTTLCache):
            with self._lock:
                self._cache.save_video(video_id, str(video_path))
                self._load_queue.delete(video_id)
        else:
            cap = cv2.VideoCapture(str(video_path))
            frame_index = 0
            total_frames = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
            while cap.isOpened():
                while self._frame_name(video_id, frame_index) in self._cache:
                    frame_index += 1
                if frame_index >= total_frames:
                    break
                cap.set(cv2.CAP_PROP_POS_FRAMES, frame_index)
                ret, frame = cap.read()
                if not ret:
                    break
                frame = np.array(frame)
                self.add_frame_to_cache(frame, video_id, frame_index)
                frame_index = int(cap.get(cv2.CAP_PROP_POS_FRAMES))
            cap.release()

    def download_video(self, api: sly.Api, video_id: int, **kwargs) -> None:
        """
        Download video if needed and add it to cache. If video is already in cache, do nothing.
        If "return_images" in kwargs and is True, returns list of frames.
        """
        return_images = kwargs.get("return_images", True)
        progress_cb = kwargs.get("progress_cb", None)

        video_info = api.video.get_info_by_id(video_id)
        self._wait_if_in_queue(video_id, api.logger)
        if not video_id in self._cache:
            self._load_queue.set(video_id, video_id)
            sly.logger.debug("Downloading video #%s", video_id)
            temp_video_path = Path("/tmp/smart_cache").joinpath(
                f"_{sly.rand_str(6)}_" + video_info.name
            )
            api.video.download_path(video_id, temp_video_path, progress_cb=progress_cb)
            self.add_video_to_cache(video_id, temp_video_path)
        if return_images:
            return self.get_frames_from_cache(video_id, list(range(video_info.frames_count)))

    def add_cache_endpoint(self, server: FastAPI):
        @server.post("/smart_cache")
        def cache_endpoint(request: Request, task: BackgroundTasks):
            task.add_task(
                self.cache_task,
                api=request.state.api,
                state=request.state.state,
            )
            return {"message": "Cache task started."}

    def add_cache_files_endpoint(self, server: FastAPI):
        @server.post("/smart_cache_files")
        def cache_files_endpoint(
            request: Request,
            task: BackgroundTasks,
            files: List[UploadFile],
            settings: str = Form("{}"),
        ):
            state = json.loads(settings)
            task.add_task(
                self.cache_files_task,
                files=files,
                state=state,
            )
            return {"message": "Cache task started."}

    def cache_task(self, api: sly.Api, state: dict):
        if "server_address" in state and "api_token" in state:
            api = sly.Api(state["server_address"], state["api_token"])
        api.logger.debug("Request state in cache endpoint", extra=state)
        image_ids, task_type = self._parse_state(state)
        kwargs = {"return_images": False}

        if task_type is InferenceImageCache._LoadType.ImageId:
            if "dataset_id" in state:
                self.download_images(api, state["dataset_id"], image_ids, **kwargs)
            else:
                for img_id in image_ids:
                    self.download_image(api, img_id)
        elif task_type is InferenceImageCache._LoadType.ImageHash:
            self.download_images_by_hashes(api, image_ids, **kwargs)
        elif task_type is InferenceImageCache._LoadType.Frame:
            video_id = state["video_id"]
            self.download_frames(api, video_id, image_ids, **kwargs)
        elif task_type is InferenceImageCache._LoadType.Video:
            video_id = image_ids
            self.download_video(api, video_id, **kwargs)

    def add_frame_to_cache(self, frame: np.ndarray, video_id: int, frame_index: int):
        name = self._frame_name(video_id, frame_index)
        self._add_to_cache(name, frame)

    def cache_files_task(self, files: List[UploadFile], state: dict):
        sly.logger.debug("Request state in cache endpoint", extra=state)
        image_ids, task_type = self._parse_state(state)

        if task_type in (
            InferenceImageCache._LoadType.ImageId,
            InferenceImageCache._LoadType.ImageHash,
        ):
            for image_id, file in zip(image_ids, files):
                image = sly.image.read_bytes(file.file.read())
                name = self._image_name(image_id)
                self._add_to_cache(name, image)
        elif task_type is InferenceImageCache._LoadType.Frame:
            video_id = state["video_id"]
            for frame_index, file in zip(image_ids, files):
                frame = sly.image.read_bytes(file.file.read())
                self.add_frame_to_cache(frame, video_id, frame_index)
        elif task_type is InferenceImageCache._LoadType.Video:
            video_id = image_ids
            temp_video_path = Path("/tmp/smart_cache").joinpath(
                f"_{sly.rand_str(6)}_" + files[0].file.name
            )
            with open(temp_video_path, "wb") as f:
                shutil.copyfileobj(files[0].file, f)
            self._wait_if_in_queue(video_id, sly.logger)
            self._load_queue.set(video_id, video_id)
            self.add_video_to_cache(video_id, str(temp_video_path))

    def run_cache_task_manually(
        self,
        api: sly.Api,
        list_of_ids_ranges_or_hashes: List[Union[str, int, List[int]]],
        *,
        dataset_id: Optional[int] = None,
        video_id: Optional[int] = None,
    ) -> None:
        """
        Run cache_task in new thread.

        :param api: supervisely api
        :type api: sly.Api
        :param list_of_ids_ranges_or_hashes: information abount images/frames need to be loaded;
        to download images, pass list of integer IDs (`dataset_id` requires)
        or list of hash strings (`dataset_id` could be None);
        to download frames, pass list of pairs of indices of the first and last frame
        and `video_id` (ex.: [[1, 3], [5, 5], [7, 10]]);
        to download video, pass None and `video_id` (only for persistent cache)
        :type list_of_ids_ranges_or_hashes: List[Union[str, int, List[int]]]
        :param dataset_id: id of dataset on supervisely platform; default is None
        :type dataset_id: Optional[int]
        :param video_id: id of video on supervisely platform; default is None
        :type video_id: Optional[int]
        """
        state = {}
        if list_of_ids_ranges_or_hashes is None:
            api.logger.debug("Got a task to add video to cache")
            if not isinstance(self._cache, PersistentImageTTLCache):
                raise ValueError("Video can be added only to persistent cache")
            state["video_id"] = video_id
        elif isinstance(list_of_ids_ranges_or_hashes[0], str):
            api.logger.debug("Got a task to add images using hash")
            state["image_hashes"] = list_of_ids_ranges_or_hashes
        elif video_id is None:
            if dataset_id is None:
                api.logger.error("dataset_id or video_id must be defined if not hashes are used")
                return
            api.logger.debug("Got a task to add images using IDs")
            state["image_ids"] = list_of_ids_ranges_or_hashes
            state["dataset_id"] = dataset_id
        else:
            api.logger.debug("Got a task to add frames")
            state["video_id"] = video_id
            state["frame_ranges"] = list_of_ids_ranges_or_hashes

        thread = Thread(target=self.cache_task, kwargs={"api": api, "state": state})
        thread.start()

    def set_project_meta(self, project_id, project_meta):
        pr_meta_name = self._project_meta_name(project_id)
        if isinstance(self._cache, PersistentImageTTLCache):
            self._cache.save_project_meta(pr_meta_name, project_meta)
        else:
            self._cache[pr_meta_name] = project_meta

    def get_project_meta(self, api: sly.Api, project_id: int):
        pr_meta_name = self._project_meta_name(project_id)
        if isinstance(self._cache, PersistentImageTTLCache):
            if pr_meta_name in self._cache:
                return self._cache.get_project_meta(pr_meta_name)
            project_meta = sly.ProjectMeta.from_json(api.project.get_meta(project_id))
            self._cache.save_project_meta(pr_meta_name, project_meta)
            return project_meta
        else:
            if pr_meta_name in self._cache:
                return self._cache[pr_meta_name]
            project_meta = sly.ProjectMeta.from_json(api.project.get_meta(project_id))
            self._cache[pr_meta_name] = project_meta
            return project_meta

    @property
    def ttl(self):
        return self._ttl

    @property
    def tmp_path(self):
        if self.is_persistent:
            return str(self._data_dir)
        return None

    def _parse_state(self, state: dict) -> Tuple[List[Any], _LoadType]:
        if "image_ids" in state:
            return state["image_ids"], InferenceImageCache._LoadType.ImageId
        elif "image_hashes" in state:
            return state["image_hashes"], InferenceImageCache._LoadType.ImageHash
        elif "video_id" in state:
            if "frame_ranges" in state:
                frame_ranges = state["frame_ranges"]
                frames = []
                for fr_range in frame_ranges:
                    shift = 1
                    if fr_range[0] > fr_range[1]:
                        shift = -1
                    start, end = fr_range[0], fr_range[1] + shift
                    frames.extend(list(range(start, end, shift)))
                return frames, InferenceImageCache._LoadType.Frame
            elif "frame_indexes" in state:
                return state["frame_indexes"], InferenceImageCache._LoadType.Frame
            return state["video_id"], InferenceImageCache._LoadType.Video
        raise ValueError("State has no proper fields: image_ids, image_hashes or video_id")

    def _add_to_cache(
        self,
        names: Union[str, List[str]],
        images: Union[np.ndarray, List[np.ndarray]],
    ):
        if isinstance(names, str):
            names = [names]

        if isinstance(images, np.ndarray):
            images = [images]

        if len(images) != len(names):
            raise ValueError(
                f"Number of images and names do not match: {len(images)} != {len(names)}"
            )

        for name, img in zip(names, images):
            with self._lock:
                self._cache.save_image(name, img)
                self._load_queue.delete(name)

    def _image_name(self, id_or_hash: Union[str, int]) -> str:
        if isinstance(id_or_hash, int):
            return f"image_{id_or_hash}"
        hash_wo_slash = id_or_hash.replace("/", "-")
        return f"image_{hash_wo_slash}"

    def _frame_name(self, video_id: int, frame_index: int) -> str:
        return f"frame_{video_id}_{frame_index}"

    def _video_name(self, video_id: int, video_name: str) -> str:
        return f"video_{video_id}.{video_name.split('.')[-1]}"

    def _project_meta_name(self, project_id: int) -> str:
        return f"project_meta_{project_id}"

    def _download_many(
        self,
        indexes: List[Union[int, str]],
        name_cunstructor: Callable[[int], str],
        load_generator: Callable[
            [List[int]],
            Generator[Tuple[Union[int, str], np.ndarray], None, None],
        ],
        logger: Logger,
        return_images: bool = True,
    ) -> Optional[List[np.ndarray]]:
        indexes_to_load = []
        pos_by_name = {}
        all_frames = [None for _ in range(len(indexes))]

        for pos, hash_or_id in enumerate(indexes):
            name = name_cunstructor(hash_or_id)
            self._wait_if_in_queue(name, logger)

            if name not in self._cache:
                self._load_queue.set(name, hash_or_id)
                indexes_to_load.append(hash_or_id)
                pos_by_name[name] = pos
            elif return_images is True:
                all_frames[pos] = self._cache.get_image(name)

        if len(indexes_to_load) > 0:
            for id_or_hash, image in load_generator(indexes_to_load):
                name = name_cunstructor(id_or_hash)
                self._add_to_cache(name, image)

                if return_images:
                    pos = pos_by_name[name]
                    all_frames[pos] = image

        logger.debug(f"All stored files: {sorted(os.listdir(self.tmp_path))}")
        logger.debug(f"Images/Frames added to cache: {indexes_to_load}")
        logger.debug(f"Images/Frames found in cache: {set(indexes).difference(indexes_to_load)}")

        if return_images:
            return all_frames
        return

    def _wait_if_in_queue(self, name, logger: Logger):
        if name in self._load_queue:
            logger.debug(f"Waiting for other task to load {name}")

        while name in self._load_queue:
            # TODO: sleep if slowdown
            sleep(0.1)
            continue



import functools
import json
import time
from queue import Queue
from threading import Event, Thread
from typing import Any, Dict, List, Optional, Union

import numpy as np
from fastapi import BackgroundTasks, Form, Request, UploadFile

import supervisely as sly
import supervisely.nn.inference.tracking.functional as F
from supervisely.annotation.label import Geometry, Label
from supervisely.nn.inference import Inference
from supervisely.nn.inference.tracking.tracker_interface import TrackerInterface
from supervisely.nn.prediction_dto import PredictionSegmentation


class MaskTracking(Inference):
    def __init__(
        self,
        model_dir: Optional[str] = None,
        custom_inference_settings: Optional[Union[Dict[str, Any], str]] = None,
    ):
        Inference.__init__(
            self,
            model_dir,
            custom_inference_settings,
            sliding_window_mode=None,
            use_gui=False,
        )

        try:
            self.load_on_device(model_dir, "cuda")
        except RuntimeError:
            self.load_on_device(model_dir, "cpu")
            sly.logger.warn("Failed to load model on CUDA device.")

        sly.logger.debug(
            "Smart cache params",
            extra={
                "ttl": sly.env.smart_cache_ttl(),
                "maxsize": sly.env.smart_cache_size(),
            },
        )

    def _deserialize_geometry(self, data: dict):
        geometry_type_str = data["type"]
        geometry_json = data["data"]
        return sly.deserialize_geometry(geometry_type_str, geometry_json)

    def get_info(self):
        info = super().get_info()
        info["task type"] = "tracking"
        return info

    def _track_api(self, api: sly.Api, context: dict):
        # unused fields:
        context["trackId"] = "auto"
        context["objectIds"] = []
        context["figureIds"] = []
        if "direction" not in context:
            context["direction"] = "forward"

        input_geometries: list = context["input_geometries"]

        self.video_interface = TrackerInterface(
            context=context,
            api=api,
            load_all_frames=True,
            notify_in_predict=True,
            per_point_polygon_tracking=False,
            frame_loader=self.cache.download_frame,
            frames_loader=self.cache.download_frames,
            should_notify=False,
        )

        range_of_frames = [
            self.video_interface.frames_indexes[0],
            self.video_interface.frames_indexes[-1],
        ]

        if self.cache.is_persistent:
            # if cache is persistent, run cache task for whole video
            self.cache.run_cache_task_manually(
                api,
                None,
                video_id=self.video_interface.video_id,
            )
        else:
            # if cache is not persistent, run cache task for range of frames
            self.cache.run_cache_task_manually(
                api,
                [range_of_frames],
                video_id=self.video_interface.video_id,
            )

        api.logger.info("Starting tracking process")
        # load frames
        frames = self.video_interface.frames
        # combine several binary masks into one multilabel mask
        i = 0
        label2id = {}

        for input_geom in input_geometries:
            geometry = self._deserialize_geometry(input_geom)
            if not isinstance(geometry, sly.Bitmap) and not isinstance(geometry, sly.Polygon):
                raise TypeError(f"This app does not support {geometry.geometry_name()} tracking")
            # convert polygon to bitmap
            if isinstance(geometry, sly.Polygon):
                polygon_obj_class = sly.ObjClass("polygon", sly.Polygon)
                polygon_label = sly.Label(geometry, polygon_obj_class)
                bitmap_obj_class = sly.ObjClass("bitmap", sly.Bitmap)
                bitmap_label = polygon_label.convert(bitmap_obj_class)[0]
                geometry = bitmap_label.geometry
            if i == 0:
                multilabel_mask = geometry.data.astype(int)
                multilabel_mask = np.zeros(frames[0].shape, dtype=np.uint8)
                geometry.draw(bitmap=multilabel_mask, color=[1, 1, 1])
                i += 1
            else:
                i += 1
                geometry.draw(bitmap=multilabel_mask, color=[i, i, i])
            label2id[i] = {
                "original_geometry": geometry.geometry_name(),
            }

        # run tracker
        tracked_multilabel_masks = self.predict(frames=frames, input_mask=multilabel_mask[:, :, 0])
        tracked_multilabel_masks = np.array(tracked_multilabel_masks)

        predictions = []
        # decompose multilabel masks into binary masks
        for i in np.unique(tracked_multilabel_masks):
            if i != 0:
                predictions_for_label = []
                binary_masks = tracked_multilabel_masks == i
                geometry_type = label2id[i]["original_geometry"]
                for j, mask in enumerate(binary_masks[1:]):
                    # check if mask is not empty
                    if not np.any(mask):
                        api.logger.info(f"Empty mask on frame {context['frameIndex'] + j + 1}")
                        predictions_for_label.append(None)
                    else:
                        # frame_idx = j + 1
                        geometry = sly.Bitmap(mask)
                        predictions_for_label.append(
                            {"type": geometry.geometry_name(), "data": geometry.to_json()}
                        )
                predictions.append(predictions_for_label)

        # predictions must be NxK masks: N=number of frames, K=number of objects
        predictions = list(map(list, zip(*predictions)))
        return predictions

    def _inference(self, frames: List[np.ndarray], geometries: List[Geometry]):
        results = [[] for _ in range(len(frames) - 1)]
        for geometry in geometries:
            if not isinstance(geometry, sly.Bitmap) and not isinstance(geometry, sly.Polygon):
                raise TypeError(f"This app does not support {geometry.geometry_name()} tracking")

            # combine several binary masks into one multilabel mask
            i = 0
            label2id = {}

            original_geometry = geometry.clone()

            # convert polygon to bitmap
            if isinstance(geometry, sly.Polygon):
                polygon_obj_class = sly.ObjClass("polygon", sly.Polygon)
                polygon_label = sly.Label(geometry, polygon_obj_class)
                bitmap_obj_class = sly.ObjClass("bitmap", sly.Bitmap)
                bitmap_label = polygon_label.convert(bitmap_obj_class)[0]
                geometry = bitmap_label.geometry
            if i == 0:
                multilabel_mask = geometry.data.astype(int)
                multilabel_mask = np.zeros(frames[0].shape, dtype=np.uint8)
                geometry.draw(bitmap=multilabel_mask, color=[1, 1, 1])
                i += 1
            else:
                i += 1
                geometry.draw(bitmap=multilabel_mask, color=[i, i, i])
            label2id[i] = {
                "original_geometry": original_geometry.geometry_name(),
            }
            # run tracker
            tracked_multilabel_masks = self.predict(
                frames=frames, input_mask=multilabel_mask[:, :, 0]
            )
            tracked_multilabel_masks = np.array(tracked_multilabel_masks)
            # decompose multilabel masks into binary masks
            for i in np.unique(tracked_multilabel_masks):
                if i != 0:
                    binary_masks = tracked_multilabel_masks == i
                    geometry_type = label2id[i]["original_geometry"]
                    for j, mask in enumerate(binary_masks[1:]):
                        # check if mask is not empty
                        if np.any(mask):
                            if geometry_type == "polygon":
                                bitmap_geometry = sly.Bitmap(mask)
                                bitmap_obj_class = sly.ObjClass("bitmap", sly.Bitmap)
                                bitmap_label = sly.Label(bitmap_geometry, bitmap_obj_class)
                                polygon_obj_class = sly.ObjClass("polygon", sly.Polygon)
                                polygon_labels = bitmap_label.convert(polygon_obj_class)
                                geometries = [label.geometry for label in polygon_labels]
                            else:
                                geometries = [sly.Bitmap(mask)]
                            results[j].extend(
                                [
                                    {"type": geom.geometry_name(), "data": geom.to_json()}
                                    for geom in geometries
                                ]
                            )
        return results

    def _track_api_cached(self, request: Request, context: dict):
        sly.logger.info(f"Start tracking with settings: {context}.")
        video_id = context["video_id"]
        frame_indexes = list(
            range(context["frame_index"], context["frame_index"] + context["frames"] + 1)
        )
        geometries = map(self._deserialize_geometry, context["input_geometries"])
        frames = self.cache.get_frames_from_cache(video_id, frame_indexes)
        return self._inference(frames, geometries)

    def _track_api_files(
        self, request: Request, files: List[UploadFile], settings: str = Form("{}")
    ):
        state = json.loads(settings)
        sly.logger.info(f"Start tracking with settings: {state}.")
        video_id = state["video_id"]
        frame_indexes = list(
            range(state["frame_index"], state["frame_index"] + state["frames"] + 1)
        )
        geometries = map(self._deserialize_geometry, state["input_geometries"])
        frames = []
        for file, frame_idx in zip(files, frame_indexes):
            img_bytes = file.file.read()
            frame = sly.image.read_bytes(img_bytes)
            frames.append(frame)
        sly.logger.info("Start tracking.")
        return self._inference(frames, geometries)

    def serve(self):
        super().serve()
        server = self._app.get_server()
        self.cache.add_cache_endpoint(server)
        self.cache.add_cache_files_endpoint(server)

        @server.post("/track")
        def start_track(request: Request, task: BackgroundTasks):
            task.add_task(track, request)
            return {"message": "Track task started."}

        @server.post("/track-api")
        def track_api(request: Request):
            return self._track_api(request.state.api, request.state.context)

        @server.post("/track-api-files")
        def track_api_files(
            request: Request,
            files: List[UploadFile],
            settings: str = Form("{}"),
        ):
            return self._track_api_files(request, files, settings)

        def send_error_data(func):
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                value = None
                try:
                    value = func(*args, **kwargs)
                except Exception as exc:
                    # print("An error occured:")
                    # print(traceback.format_exc())
                    request: Request = args[0]
                    context = request.state.context
                    api: sly.Api = request.state.api
                    track_id = context["trackId"]
                    api.logger.error(f"An error occured: {repr(exc)}")

                    api.post(
                        "videos.notify-annotation-tool",
                        data={
                            "type": "videos:tracking-error",
                            "data": {
                                "trackId": track_id,
                                "error": {"message": repr(exc)},
                            },
                        },
                    )
                return value

            return wrapper

        @send_error_data
        def track(request: Request):
            context = request.state.context
            api: sly.Api = request.state.api

            self.video_interface = TrackerInterface(
                context=context,
                api=api,
                load_all_frames=True,
                notify_in_predict=True,
                per_point_polygon_tracking=False,
                frame_loader=self.cache.download_frame,
                frames_loader=self.cache.download_frames,
            )
            api.logger.info("Starting tracking process")
            # load frames
            frames = self.video_interface.frames
            # combine several binary masks into one multilabel mask
            i = 0
            label2id = {}

            def _upload_loop(q: Queue, stop_event: Event, video_interface: TrackerInterface):
                try:
                    while True:
                        items = []
                        while not q.empty():
                            items.append(q.get_nowait())
                        if len(items) > 0:
                            video_interface.add_object_geometries_on_frames(*list(zip(*items)))
                        if stop_event.is_set():
                            video_interface._notify(True, task="stop tracking")
                            return
                        time.sleep(1)
                except Exception as e:
                    api.logger.error("Error in upload loop: %s", str(e), exc_info=True)
                    video_interface._notify(True, task="stop tracking")
                    video_interface.global_stop_indicatior = True
                    raise

            upload_queue = Queue()
            stop_upload_event = Event()
            Thread(
                target=_upload_loop,
                args=[upload_queue, stop_upload_event, self.video_interface],
                daemon=True,
            ).start()

            try:
                for (fig_id, geometry), obj_id in zip(
                    self.video_interface.geometries.items(),
                    self.video_interface.object_ids,
                ):
                    original_geometry = geometry.clone()
                    if not isinstance(geometry, sly.Bitmap) and not isinstance(
                        geometry, sly.Polygon
                    ):
                        stop_upload_event.set()
                        raise TypeError(
                            f"This app does not support {geometry.geometry_name()} tracking"
                        )
                    # convert polygon to bitmap
                    if isinstance(geometry, sly.Polygon):
                        polygon_obj_class = sly.ObjClass("polygon", sly.Polygon)
                        polygon_label = sly.Label(geometry, polygon_obj_class)
                        bitmap_obj_class = sly.ObjClass("bitmap", sly.Bitmap)
                        bitmap_label = polygon_label.convert(bitmap_obj_class)[0]
                        geometry = bitmap_label.geometry
                    if i == 0:
                        multilabel_mask = geometry.data.astype(int)
                        multilabel_mask = np.zeros(frames[0].shape, dtype=np.uint8)
                        geometry.draw(bitmap=multilabel_mask, color=[1, 1, 1])
                        i += 1
                    else:
                        i += 1
                        geometry.draw(bitmap=multilabel_mask, color=[i, i, i])
                    label2id[i] = {
                        "fig_id": fig_id,
                        "obj_id": obj_id,
                        "original_geometry": original_geometry.geometry_name(),
                    }
                # run tracker
                tracked_multilabel_masks = self.predict(
                    frames=frames, input_mask=multilabel_mask[:, :, 0]
                )
                tracked_multilabel_masks = np.array(tracked_multilabel_masks)
                # decompose multilabel masks into binary masks
                for i in np.unique(tracked_multilabel_masks):
                    if i != 0:
                        binary_masks = tracked_multilabel_masks == i
                        fig_id = label2id[i]["fig_id"]
                        obj_id = label2id[i]["obj_id"]
                        geometry_type = label2id[i]["original_geometry"]
                        for j, mask in enumerate(binary_masks[1:]):
                            # check if mask is not empty
                            if not np.any(mask):
                                api.logger.info(
                                    f"Skipping empty mask on frame {self.video_interface.frame_index + j + 1}"
                                )
                                # update progress bar anyway (otherwise it will not be finished)
                                self.video_interface._notify(task="add geometry on frame")
                            else:
                                if geometry_type == "polygon":
                                    bitmap_geometry = sly.Bitmap(mask)
                                    bitmap_obj_class = sly.ObjClass("bitmap", sly.Bitmap)
                                    bitmap_label = sly.Label(bitmap_geometry, bitmap_obj_class)
                                    polygon_obj_class = sly.ObjClass("polygon", sly.Polygon)
                                    polygon_labels = bitmap_label.convert(polygon_obj_class)
                                    geometries = [label.geometry for label in polygon_labels]
                                else:
                                    geometries = [sly.Bitmap(mask)]
                                for l, geometry in enumerate(geometries):
                                    if l == len(geometries) - 1:
                                        notify = True
                                    else:
                                        notify = False
                                    upload_queue.put(
                                        (
                                            geometry,
                                            obj_id,
                                            self.video_interface.frames_indexes[j + 1],
                                            notify,
                                        )
                                    )
                        if self.video_interface.global_stop_indicatior:
                            stop_upload_event.set()
                            return
                        api.logger.info(f"Figure with id {fig_id} was successfully tracked")
            except Exception:
                stop_upload_event.set()
                raise
            stop_upload_event.set()



from supervisely.app import DataJson, StateJson
from supervisely.app.widgets import Widget

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


class Input(Widget):
    class Routes:
        VALUE_CHANGED = "value_changed"

    def __init__(
        self,
        value: str = "",
        minlength: int = 0,
        maxlength: int = 1000,
        placeholder: str = "",
        size: Literal["mini", "small", "large"] = None,
        readonly: bool = False,
        type: Literal["text", "password", "textarea"] = "text",
        widget_id: str = None,
        icon: Literal["search", "edit"] = None,
        autosize_textarea: bool = True,
    ):
        self._value = value  # initial value
        self._minlength = minlength
        self._maxlength = maxlength
        self._placeholder = placeholder
        self._size = size
        self._readonly = readonly
        self._widget_id = widget_id
        self._changes_handled = False
        self._type = type
        self._icon = icon
        self._autosize_textarea = autosize_textarea

        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self):
        return {
            "minlength": self._minlength,
            "maxlength": self._maxlength,
            "placeholder": self._placeholder,
            "size": self._size,
            "readonly": self._readonly,
            "autosize": self._autosize_textarea,
        }

    def get_json_state(self):
        return {
            "value": self._value,
            "type": self._type,
            "icon": self._icon,
        }

    @property
    def autosize_textarea(self):
        return self._autosize_textarea

    @autosize_textarea.setter
    def autosize_textarea(self, value: bool):
        self._autosize_textarea = value
        DataJson()[self.widget_id]["autosize_textarea"] = value

    def set_value(self, value):
        StateJson()[self.widget_id]["value"] = value
        StateJson().send_changes()

    def get_value(self):
        return StateJson()[self.widget_id]["value"]

    def set_type(self, value: Literal["text", "password"]):
        StateJson()[self.widget_id]["type"] = value
        StateJson().send_changes()

    def get_type(self):
        return StateJson()[self.widget_id]["type"]

    def set_icon(self, value: Literal["search", "edit"]):
        StateJson()[self.widget_id]["icon"] = value
        StateJson().send_changes()

    def get_icon(self):
        return StateJson()[self.widget_id]["icon"]

    def is_readonly(self):
        return DataJson()[self.widget_id]["readonly"]

    def enable_readonly(self):
        DataJson()[self.widget_id]["readonly"] = True
        DataJson().send_changes()

    def disable_readonly(self):
        DataJson()[self.widget_id]["readonly"] = False
        DataJson().send_changes()

    def value_changed(self, func):
        route_path = self.get_route_path(Input.Routes.VALUE_CHANGED)
        server = self._sly_app.get_server()
        self._changes_handled = True

        @server.post(route_path)
        def _click():
            res = self.get_value()
            # self._value = res  # commented cause we do not need to update initial value
            func(res)

        return _click



import os
import shutil
import stat
import subprocess
from typing import Optional

import requests

from supervisely.api.api import Api
from supervisely.io.fs import mkdir
from supervisely.sly_logger import logger

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

VPN_CONFIGURATION_DIR = "~/supervisely-network"


def supervisely_vpn_network(
    action: Optional[Literal["up", "down"]] = "up",
    raise_on_error: Optional[bool] = True,
) -> None:
    """Connects to Supervisely network using WireGuard VPN.
    Reads Supervisely API settings from the environment variables.

    :param action: The action to perform, either "up" or "down".
    :type action: Optional[Literal["up", "down"]]
    :param raise_on_error: If True, an exception will be raised if an error occurs.
    :type raise_on_error: Optional[bool]
    :raises RuntimeError: If wg-quick is not available in the system and raise_on_error is True.
    :raises subprocess.CalledProcessError: If an error occurs while connecting and raise_on_error is True.
    :raises RuntimeError: If an error occurs while connecting and raise_on_error is True.
    """
    if shutil.which("wg-quick") is None:
        if raise_on_error:
            raise RuntimeError(
                "wg-quick is not available in the system. "
                "Please refer to this documentation to install required packages: "
                "https://developer.supervisely.com/app-development/advanced/advanced-debugging#prepare-environment"
            )
        else:
            logger.error(
                "wg-quick is not available in the system. "
                "Please refer to this documentation to install required packages: "
                "https://developer.supervisely.com/app-development/advanced/advanced-debugging#prepare-environment"
            )
            return
    else:
        logger.info("wg-quick is available in the system, will try to connect to VPN.")

    logger.info("wg-quick reqires root privileges, you may be asked to enter your password.")
    network_dir = os.path.expanduser(VPN_CONFIGURATION_DIR)
    mkdir(network_dir)
    os.chdir(network_dir)

    config_file = os.path.join(network_dir, "wg0.conf")
    public_key = os.path.join(network_dir, "public.key")
    private_key = os.path.join(network_dir, "private.key")

    logger.info(f"VPN configuration directory: {network_dir}")
    api = Api()

    # Try to bring down the connection if it's active.
    logger.info("Trying to bring down the connection...")
    try:
        subprocess.run(
            ["wg-quick", "down", config_file],
            check=False,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        logger.info("Connection has been brought down successfully.")
    except subprocess.CalledProcessError as e:
        logger.info(f"The connection was not active: {e.stderr.decode()}.")

    # Generate the private and public keys if they don't exist.
    if not os.path.exists(public_key):
        logger.info(f"Public key not found in {public_key}, generating a new one...")
        with open(private_key, "w") as priv_key_file:
            subprocess.run(["wg", "genkey"], stdout=priv_key_file)
        os.chmod(private_key, stat.S_IRUSR | stat.S_IWUSR)  # Change file permissions to 600
        with open(private_key) as priv_key_file, open(public_key, "w") as pub_key_file:
            subprocess.run(["wg", "pubkey"], stdin=priv_key_file, stdout=pub_key_file)
    else:
        logger.info(f"Public key found in {public_key}, using it...")

    # Register the connection with the server.
    logger.info(f"Registering the connection with the server: {api.server_address}...")
    response = requests.post(
        f"{api.server_address}/net/register/{api.token}/{open(public_key).read().strip()}",
        timeout=5,
    )
    response.raise_for_status()
    logger.info(f"Connection registered with the server, status: {response.status_code}.")
    ip, server_public_key, server_endpoint, subnet_base, gateway = response.text.split(";")

    # Update the configuration file with the server's parameters.
    logger.info(
        f"Updating wg0.conf with the following parameters: IP: {ip} "
        f"Server Public Key: {server_public_key}, Server endpoint: {server_endpoint} "
        f"Subnet base: {subnet_base}, Gateway: {gateway}"
    )
    with open(config_file, "w") as f:
        f.write(
            f"""
            [Interface]
            PrivateKey = {open(private_key).read().strip()}
            Address = {ip}/16
            [Peer]
            PublicKey = {server_public_key}
            AllowedIPs = {subnet_base}
            Endpoint = {server_endpoint}
            PersistentKeepalive = 25
            """
        )
    os.chmod(config_file, 0o600)

    logger.info("Configuration saved, bringing up the connection...")
    try:
        subprocess.run(
            ["wg-quick", action, config_file],
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        logger.info(
            "WireGuard interface has been successfully brought up, checking the connection..."
        )
    except subprocess.CalledProcessError as e:
        if raise_on_error:
            raise
        else:
            logger.warning(f"Error while connecting to VPN, try again. Error: {e.stderr.decode()}")
            return

    # Check the connection.
    test_response = requests.get(f"http://{gateway}", timeout=5)
    test_response.raise_for_status()

    if not test_response.ok:
        logger.warning(f"Error while connecting to VPN, try again. Error: {test_response.text}")
        if raise_on_error:
            raise RuntimeError(f"Error while connecting to VPN: {test_response.text}")
    else:
        logger.info(f"VPN connection has been successfully established to {gateway}")


def create_debug_task(team_id, port="8000"):
    api = Api()
    me = api.user.get_my_info()
    session_name = me.login + "-development"
    module_id = api.app.get_ecosystem_module_id("supervisely-ecosystem/while-true-script-v2")
    sessions = api.app.get_sessions(team_id, module_id, session_name=session_name)
    redirect_requests = {"token": api.token, "port": port}
    task = None
    for session in sessions:
        if (session.details["meta"].get("redirectRequests") == redirect_requests) and (
            session.details["status"] == str(api.app.Status.QUEUED)
        ):
            task = session.details
            if "id" not in task:
                task["id"] = task["taskId"]
            logger.info(f"Debug task already exists: {task['id']}")
            break
    workspaces = api.workspace.get_list(team_id)
    if task is None:
        task = api.task.start(
            agent_id=None,
            module_id=module_id,
            workspace_id=workspaces[0].id,
            task_name=session_name,
            redirect_requests=redirect_requests,
            proxy_keep_url=False,  # to ignore /net/<token>/endpoint
        )
        if type(task) is list:
            task = task[0]
        logger.info(f"Debug task has been successfully created: {task['taskId']}")
    return task



from typing import Union, NamedTuple, List, Dict
from supervisely.app.widgets.apexchart.apexchart import Apexchart
from supervisely.app.content import StateJson, DataJson

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

"""
series = [
    {"name": "Team A", "data": 44},
    {"name": "Team B", "data": 55},
    {"name": "Team C", "data": 13},
    {"name": "Team D", "data": 43},
    {"name": "Team E", "data": 22},
]

pie_chart = sly.app.widgets.PieChart(
    title="Simple Pie",
    series=series,
    stroke_width=1,
    data_labels=True,
    height=350,
    type="pie",
)

text = sly.app.widgets.Text("Click on the slice to see it's value", status="info")
container = sly.app.widgets.Container(widgets=[pie_chart, text])
card = sly.app.widgets.Card(title="Pie Chart", content=container)
app = sly.Application(layout=card)


@pie_chart.click
def show_selection(datapoint: sly.app.widgets.PieChart.ClickedDataPoint):
    data_name = datapoint.data["name"]
    data_value = datapoint.data["value"]
    data_index = datapoint.data_index
    text.set(f"Selected slice: {data_name}, Value: {data_value}, Index: {data_index}", "info")
"""


class PieChart(Apexchart):
    class ClickedDataPoint(NamedTuple):
        """Class, representing clicked datapoint, which contains information about series, data index and data itself.
        It will be returned after click event on datapoint in immutable namedtuple
        with fields: series_index, data_index, data."""

        series_index: int
        data_index: int
        data: dict

    def __init__(
        self,
        title: str,
        series: List[Dict[str, Union[int, float]]] = [],
        stroke_width: int = 2,
        data_labels: bool = False,
        height: Union[int, str] = 350,
        type: Literal["pie", "donut"] = "pie",
    ):
        self._title = title
        self._series = series
        self._stroke_width = stroke_width
        self._data_labels = data_labels
        self._widget_height = height

        self._series_labels = [serie["name"] for serie in self._series]
        self._series_data = [serie["data"] for serie in self._series]

        if type != "pie" and type != "donut":
            raise ValueError("type must be 'pie' or 'donut'")
        self._type = type

        self._options = {
            "labels": self._series_labels,
            "chart": {"type": self._type},
            "dataLabels": {"enabled": self._data_labels},
            "stroke": {"width": self._stroke_width},
            "title": {"text": self._title},
        }
        super(PieChart, self).__init__(
            series=self._series_data,
            options=self._options,
            type=self._type,
            height=self._widget_height,
        )

    def _manage_series(self, names: List[str], values: List[Union[int, float]], set: bool = False):
        """This is a private method, which should not be used directly. Use add_series() or set_series() instead.
        It will add or set series to the chart, depending on set parameter. If set is True, all previous series will be
        deleted, otherwise new series will be added to the chart.

        :param names: list of names of the slices in series, which will be displayed on the chart
        :type names: List[str]
        :param values: list of values of the slices in series, which will be displayed on the chart
        :type values: List[Union[int, float]
        :param set: if True, all previous series will be deleted, otherwise new series will be added to the chart,
            defaults to False
        :type set: bool, optional
        :raises ValueError: if names and values has different length
        :raises ValueError: if any of values is not int or float
        """
        if len(names) != len(values):
            raise ValueError(
                f"Names and values has different length: {len(names)} != {len(values)}"
            )
        for value in values:
            if not isinstance(value, (int, float)):
                raise ValueError(
                    f"All values must be ints or floats, but {value} is {type(value)}."
                )

        if set:
            self._series = values
            self._options["labels"] = names
        else:
            self._series.extend(values)
            self._options["labels"].extend(names)

        self.update_data()
        DataJson().send_changes()

    def add_series(self, names: List[str], values: List[Union[int, float]]):
        # print(self._options["yaxis"]["min"], self._options["yaxis"]["max"])
        self._manage_series(names, values)

    def set_series(self, names: List[str], values: List[Union[int, float]]):
        """Sets series to the chart, deleting all previous series. Len of names and values must be equal,
        otherwise ValueError will be raised.

        :param names: list of names of the slices in series, which will be displayed on the chart
        :type names: List[str]
        :param values: list of values of the slices in series, which will be displayed on the chart
        :type values: List[Union[int, float]]
        """
        self._manage_series(names, values, set=True)

    def get_series(self, index: int) -> Dict[str, Union[str, int, float]]:
        """Returns series by index. If index is out of range, IndexError will be raised.
        Returned series is a dict with keys "name" and "data".

        :param index: index of the series, if index is out of range, IndexError will be raised
        :type index: int
        :raises TypeError: if index is not int
        :raises IndexError: if index is out of range
        :return: series name ans data by given index
        :rtype: Dict[str, Union[str, int, float]]
        """

        if not isinstance(index, int):
            raise TypeError(f"Index must be int, but {index} is {type(index)}.")
        try:
            return {"name": self._options["labels"][index], "data": self._series[index]}
        except IndexError:
            raise IndexError(f"Series with index {index} does not exist.")

    def delete_series(self, index: int):
        """Removes series by index from the chart. If index is out of range, IndexError will be raised.

        :param index: index of the series to delete
        :type index: int
        :raises TypeError: if index is not int
        :raises IndexError: if index is out of range
        """
        if not isinstance(index, int):
            raise TypeError(f"Index must be int, but {index} is {type(index)}.")
        try:
            del self._series[index]
            del self._options["labels"][index]
        except IndexError:
            raise IndexError(f"Series with index {index} does not exist.")
        self.update_data()
        DataJson().send_changes()

    def get_clicked_value(self):
        return StateJson()[self.widget_id]["clicked_value"]

    def get_clicked_datapoint(self) -> Union[ClickedDataPoint, None]:
        """Returns clicked datapoint as a ClickedDataPoint object, which is a namedtuple with fields:
        series_index, data_index and data. If click was outside of the slices, None will be returned.

        :return: clicked datapoint as a ClickedDataPoint object or None if click was outside of the slices
        :rtype: Union[ClickedDataPoint, None]
        """
        value = self.get_clicked_value()
        series_index = value["seriesIndex"]
        data_index = value["dataPointIndex"]

        if series_index == -1 or data_index == -1:
            # If click was outside of the slices.
            return

        data = {
            "name": self._options["labels"][data_index],
            "value": self._series[data_index],
        }

        res = self.ClickedDataPoint(series_index, data_index, data)
        return res



from typing import List, Optional, Dict, Union, Literal
from supervisely.app import DataJson
from supervisely.app.widgets import Widget


class Tooltip(Widget):
    def __init__(
        self,
        text: Union[str, List[str]],
        content: Widget,
        color_theme: Optional[Literal["dark", "light"]] = "dark",
        placement: Optional[
            Literal[
                "top",
                "top-start",
                "top-end",
                "bottom",
                "bottom-start",
                "bottom-end",
                "left",
                "left-start",
                "left-end",
                "right",
                "right-start",
                "right-end",
            ]
        ] = "bottom",
        offset: Optional[int] = 0,
        transition: Optional[
            Literal[
                "el-fade-in-linear",
                "el-fade-in",
            ]
        ] = "el-fade-in-linear",
        visible_arrow: Optional[bool] = True,
        open_delay: Optional[int] = 0,
        enterable: Optional[bool] = True,
        hide_after: Optional[int] = 0,
        widget_id: Optional[str] = None,
    ):
        self._text = text
        self._content = content
        self._color_theme = color_theme
        self._placement = placement
        self._offset = offset
        self._transition = transition
        self._visible_arrow = visible_arrow
        self._open_delay = open_delay
        self._enterable = enterable
        self._hide_after = hide_after
        self._multiline = True if isinstance(self._text, List) else False

        if open_delay >= hide_after and hide_after != 0:
            raise ValueError("The value 'open_delay' must be less than 'hide_after'")

        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self) -> Dict:
        return {
            "text": self._text,
            "color_theme": self._color_theme,
            "placement": self._placement,
            "offset": self._offset,
            "transition": self._transition,
            "visible_arrow": self._visible_arrow,
            "open_delay": self._open_delay,
            "enterable": self._enterable,
            "hide_after": self._hide_after,
            "multiline": self._multiline,
        }

    def get_json_state(self):
        return None

    def set_text(self, text: Union[str, List[str]]):
        """
        To make tooltip multiline - pass text as list of lines
        """
        self._text = text
        self._multiline = True if isinstance(self._text, List) else False
        DataJson()[self.widget_id]["text"] = self._text
        DataJson()[self.widget_id]["multiline"] = self._multiline
        DataJson().send_changes()

    def set_placement(
        self,
        placement: Literal[
            "top",
            "top-start",
            "top-end",
            "bottom",
            "bottom-start",
            "bottom-end",
            "left",
            "left-start",
            "left-end",
            "right",
            "right-start",
            "right-end",
        ],
    ):
        self._placement = placement
        DataJson()[self.widget_id]["placement"] = self._placement
        DataJson().send_changes()

    def set_offset(self, offset: int):
        self._offset = offset
        DataJson()[self.widget_id]["offset"] = self._offset
        DataJson().send_changes()

    def set_transition(
        self,
        transition: Literal["el-fade-in-linear", "el-fade-in"],
    ):
        self._transition = transition
        DataJson()[self.widget_id]["transition"] = self._transition
        DataJson().send_changes()

    def set_arrow_visibility(self, visible_arrow: bool):
        self._visible_arrow = visible_arrow
        DataJson()[self.widget_id]["visible_arrow"] = self._visible_arrow
        DataJson().send_changes()

    def set_open_delay(self, open_delay: int):
        """
        Milliseconds
        """
        if open_delay >= self._hide_after and self._hide_after != 0:
            raise ValueError(
                f"The value 'open_delay: {open_delay}' must be less than 'hide_after: {self._hide_after}'"
            )
        else:
            self._open_delay = open_delay
            DataJson()[self.widget_id]["open_delay"] = self._open_delay
            DataJson().send_changes()

    def set_hide_after(self, hide_after: int):
        """
        Milliseconds
        """
        if self._open_delay >= hide_after and hide_after != 0:
            raise ValueError(
                f"The value 'hide_after: {hide_after}' must be greater than 'open_delay: {self._open_delay}'"
            )
        else:
            self._hide_after = hide_after
            DataJson()[self.widget_id]["hide_after"] = self._hide_after
            DataJson().send_changes()



import functools
import os
from typing import List, Optional, Tuple

import numpy

from supervisely._utils import rand_str as sly_rand_str
from supervisely.annotation.annotation import Annotation
from supervisely.annotation.label import Label
from supervisely.annotation.tag import Tag
from supervisely.geometry.bitmap import Bitmap
from supervisely.geometry.graph import GraphNodes, Node
from supervisely.geometry.point_location import PointLocation
from supervisely.geometry.rectangle import Rectangle
from supervisely.geometry.sliding_windows_fuzzy import (
    SlidingWindowBorderStrategy,
    SlidingWindowsFuzzy,
)
from supervisely.imaging import image as sly_image
from supervisely.io.fs import silent_remove
from supervisely.sly_logger import logger


def _process_image_path(image_path: str, rect: Rectangle) -> Tuple[str, Tuple[int, int]]:
    image = sly_image.read(image_path)
    image_size = image.shape[:2]
    image_base_dir = os.path.dirname(image_path)
    image_name, image_ext = os.path.splitext(os.path.basename(image_path))

    image_crop = sly_image.crop(image, rect)
    image_crop_path = os.path.join(
        image_base_dir, sly_rand_str(10) + "_" + image_name + "_crop" + image_ext
    )
    sly_image.write(image_crop_path, image_crop)
    return image_crop_path, image_size


def _scale_ann_to_original_size(
    ann: Annotation, original_size: Tuple[int, int], rect: Rectangle
) -> Annotation:
    updated_labels = []
    for label in ann.labels:
        updated_geometry = label.geometry
        if type(label.geometry) is Rectangle:
            updated_geometry = Rectangle(
                top=label.geometry.top + rect.top,
                left=label.geometry.left + rect.left,
                bottom=label.geometry.bottom + rect.top,
                right=label.geometry.right + rect.left,
            )

        if type(label.geometry) is Bitmap:
            bitmap_data = label.geometry.data
            bitmap_origin = PointLocation(
                label.geometry.origin.row + rect.top,
                label.geometry.origin.col + rect.left,
            )

            updated_geometry = Bitmap(data=bitmap_data, origin=bitmap_origin)

        if type(label.geometry) is GraphNodes:
            new_nodes = []
            for id, node in label.geometry.nodes.items():
                new_nodes.append(
                    Node(
                        label=id,
                        row=node.location.row + rect.top,
                        col=node.location.col + rect.left,
                    )
                )

            updated_geometry = GraphNodes(new_nodes)

        updated_labels.append(label.clone(geometry=updated_geometry))

    ann = ann.clone(img_size=original_size, labels=updated_labels)
    return ann


def _apply_agnostic_nms(labels: List[Label], iou_thres: Optional[float] = 0.5) -> List[Label]:
    # pylint: disable=import-error
    import torch

    # pylint: disable=import-error
    import torchvision

    # TODO: where we can get iou_th and conf_th?
    boxes = []
    scores = []
    for label in labels:
        label: Label
        label_rect: Rectangle = label.geometry.to_bbox()
        boxes.append(
            [
                float(label_rect.left),
                float(label_rect.top),
                float(label_rect.right),
                float(label_rect.bottom),
            ]
        )
        conf_score: Tag = label.tags.get("confidence", None)
        if conf_score is None:
            raise ValueError("Label don't have confidence score tag named 'confidence'.")
        scores.append(conf_score.value)
    if len(boxes) > 0 and len(scores) > 0:
        boxes = torch.tensor(boxes)
        scores = torch.tensor(scores)
        saved_inds = torchvision.ops.nms(boxes, scores, iou_thres)
    else:
        saved_inds = []
    saved_labels = []
    for ind in saved_inds:
        saved_labels.append(labels[ind])
    return saved_labels


def process_image_roi(func):
    """
    Decorator for processing annotation labels before and after inference.

    Crops input image before inference if kwargs['state']['rectangle_crop'] provided
    and then scales annotation back to original image size.

    Keyword arguments:

    :param image_np: Image in numpy.ndarray format (use image_path or image_np, not both)
    :type image_np: numpy.ndarray
    :param image_path: Path to image (use image_path or image_np, not both)
    :type image_path: str
    :raises: :class:`ValueError`, if image_np or image_path invalid or not provided
    :return: Annotation in json format
    :rtype: :class:`dict`
    """

    @functools.wraps(func)
    def wrapper_inference(*args, **kwargs):
        settings = kwargs["settings"]
        rectangle_json = settings.get("rectangle")

        if rectangle_json is None:
            ann = func(*args, **kwargs)
            return ann

        rectangle = Rectangle.from_json(rectangle_json)
        if "image_np" in kwargs.keys():
            image_np = kwargs["image_np"]
            if not isinstance(image_np, numpy.ndarray):
                raise ValueError("Invalid input. Image path must be numpy.ndarray")
            original_image_size = image_np.shape[:2]
            image_crop_np = sly_image.crop(image_np, rectangle)
            kwargs["image_np"] = image_crop_np
            ann = func(*args, **kwargs)
            ann = _scale_ann_to_original_size(ann, original_image_size, rectangle)
        elif "image_path" in kwargs.keys():
            image_path = kwargs["image_path"]
            if not isinstance(image_path, str):
                raise ValueError("Invalid input. Image path must be str")
            image_crop_path, original_image_size = _process_image_path(image_path, rectangle)
            kwargs["image_path"] = image_crop_path
            ann = func(*args, **kwargs)
            ann = _scale_ann_to_original_size(ann, original_image_size, rectangle)
            silent_remove(image_crop_path)
        else:
            raise ValueError("image_np or image_path not provided!")

        return ann

    return wrapper_inference


def process_image_sliding_window(func):
    @functools.wraps(func)
    def wrapper_inference(*args, **kwargs):
        settings = kwargs["settings"]
        data_to_return = kwargs["data_to_return"]
        inference_mode = settings.get("inference_mode", "full_image")
        if inference_mode != "sliding_window":
            ann = func(*args, **kwargs)
            return ann
        sliding_window_mode = settings.get("sliding_window_mode", "basic")
        if sliding_window_mode == "none":
            ann = func(*args, **kwargs)
            return ann

        sliding_window_params = settings["sliding_window_params"]
        image_path = kwargs["image_path"]
        img = sly_image.read(image_path)
        img_h, img_w = img.shape[:2]
        windowHeight = sliding_window_params.get("windowHeight", img_h)
        windowWidth = sliding_window_params.get("windowWidth", img_w)
        overlapY = sliding_window_params.get("overlapY", 0)
        overlapX = sliding_window_params.get("overlapX", 0)
        borderStrategy = sliding_window_params.get("borderStrategy", "shift_window")

        slider = SlidingWindowsFuzzy(
            [windowHeight, windowWidth], [overlapY, overlapX], borderStrategy
        )
        rectangles = []
        for window in slider.get(img.shape[:2]):
            rectangles.append(window)

        data_to_return["slides"] = []
        all_labels = []
        for rect in rectangles:
            image_crop_path, original_image_size = _process_image_path(image_path, rect)
            kwargs["image_path"] = image_crop_path
            slice_ann: Annotation = func(*args, **kwargs)
            slice_ann = _scale_ann_to_original_size(slice_ann, original_image_size, rect)
            data_to_return["slides"].append(
                {
                    "rectangle": rect.to_json(),
                    "labels": [l.to_json() for l in slice_ann.labels],
                }
            )
            all_labels.extend(slice_ann.labels)

        all_json_labels = []
        for slide in data_to_return["slides"]:
            all_json_labels.extend(slide["labels"])

        full_rect = Rectangle(0, 0, img_h, img_w)
        all_labels_slide = {"rectangle": full_rect.to_json(), "labels": all_json_labels}
        data_to_return["slides"].append(all_labels_slide)  # for visualization
        ann = Annotation.from_img_path(image_path)

        if sliding_window_mode == "advanced":
            labels_after_nms = _apply_agnostic_nms(all_labels)
            ann = ann.add_labels(labels_after_nms)
            all_labels_after_nms_slide = {
                "rectangle": full_rect.to_json(),
                "labels": [l.to_json() for l in labels_after_nms],
            }
            data_to_return["slides"].append(all_labels_after_nms_slide)
        else:
            ann = ann.add_labels(all_labels)
            data_to_return["slides"].append(all_labels_slide)
        return ann

    return wrapper_inference


def process_images_batch_roi(func):
    @functools.wraps(func)
    def wrapper_inference(*args, **kwargs):
        """
        Process images batch with ROI cropping before inference and scaling back to original size after inference.
        Pass settings with rectangle or rectangles to crop images and annotations.
        If rectangle provided, crop all images with the same rectangle.
        If rectangles provided, crop each image with corresponding rectangle.
        """
        source = kwargs["source"]
        settings = kwargs["settings"]
        if "rectangles" in settings:
            rectangles = [Rectangle.from_json(rect_json) for rect_json in settings["rectangles"]]
        elif "rectangle" in settings:
            rectangles = [Rectangle.from_json(settings["rectangle"]) for _ in source]
        else:
            return func(*args, **kwargs)

        original_images_sizes = [image_np.shape[:2] for image_np in source]
        images_crops_nps = [
            sly_image.crop(image_np, rect) for image_np, rect in zip(source, rectangles)
        ]
        kwargs["source"] = images_crops_nps
        anns = func(*args, **kwargs)
        anns = [
            _scale_ann_to_original_size(ann, original_image_size, rect)
            for ann, original_image_size, rect in zip(anns, original_images_sizes, rectangles)
        ]
        return anns

    return wrapper_inference


def process_images_batch_sliding_window(func):
    @functools.wraps(func)
    def wrapper_inference(*args, **kwargs):
        settings = kwargs["settings"]
        data_to_return = kwargs["data_to_return"]
        inference_mode = settings.get("inference_mode", "full_image")
        if inference_mode != "sliding_window":
            anns = func(*args, **kwargs)
            return anns
        sliding_window_mode = settings.get("sliding_window_mode", "basic")
        if sliding_window_mode == "none":
            anns = func(*args, **kwargs)
            return anns

        sliding_window_params = settings["sliding_window_params"]
        source = kwargs["source"]
        data_to_return["slides"] = []
        anns = []

        images_crops = []
        original_images_sizes = []
        images_slices_lengths = []
        images_shapes = []
        for src in source:
            if isinstance(src, str):
                img = sly_image.read(src)
            else:
                img = src
            img_h, img_w = img.shape[:2]
            images_shapes.append((img_h, img_w))
            windowHeight = sliding_window_params.get("windowHeight", img_h)
            windowWidth = sliding_window_params.get("windowWidth", img_w)
            overlapY = sliding_window_params.get("overlapY", 0)
            overlapX = sliding_window_params.get("overlapX", 0)
            borderStrategy = sliding_window_params.get("borderStrategy", "shift_window")

            slider = SlidingWindowsFuzzy(
                [windowHeight, windowWidth], [overlapY, overlapX], borderStrategy
            )
            rectangles = []
            for window in slider.get(img.shape[:2]):
                rectangles.append(window)

            for rect in rectangles:
                original_image_size = (img_h, img_w)
                image_crop = sly_image.crop(img, rect)
                images_crops.append(image_crop)
                original_images_sizes.append(original_image_size)
            images_slices_lengths.append(len(rectangles))

            ann = Annotation((img_h, img_w))
            anns.append(ann)

        kwargs["source"] = images_crops
        slices_anns: List[Annotation] = func(*args, **kwargs)
        slices_anns = [
            _scale_ann_to_original_size(slice_ann, original_image_size, rect)
            for slice_ann in slices_anns
        ]
        for i in range(len(images_slices_lengths)):
            i_from = sum(images_slices_lengths[:i])
            i_to = sum(images_slices_lengths[: i + 1])
            image_rects = rectangles[i_from:i_to]
            image_slices = slices_anns[i_from:i_to]
            data_to_return["slides"].append(
                [
                    {"rectangle": rect.to_json(), "labels": [l.to_json() for l in slice_ann.labels]}
                    for rect, slice_ann in zip(image_rects, image_slices)
                ]
            )

        for i, image_shape in enumerate(images_shapes):
            img_h, img_w = image_shape
            all_json_labels = [slide["labels"] for slide in data_to_return["slides"][i]]
            full_rect = Rectangle(0, 0, img_h, img_w)
            all_labels_slide = {"rectangle": full_rect.to_json(), "labels": all_json_labels}
            data_to_return["slides"][i].append(all_labels_slide)  # for visualization

            i_from = sum(images_slices_lengths[:i])
            i_to = sum(images_slices_lengths[: i + 1])
            image_slices = slices_anns[i_from:i_to]

            all_labels = []
            for slice_ann in image_slices:
                all_labels.extend(slice_ann.labels)
            if sliding_window_mode == "advanced":
                labels_after_nms = _apply_agnostic_nms(all_labels)
                anns[i] = anns[i].add_labels(labels_after_nms)
                all_labels_after_nms_slide = {
                    "rectangle": full_rect.to_json(),
                    "labels": [l.to_json() for l in labels_after_nms],
                }
                data_to_return["slides"][i].append(all_labels_after_nms_slide)
            else:
                anns[i] = anns[i].add_labels(all_labels)
                data_to_return["slides"][i].append(all_labels_slide)

        return anns

    return wrapper_inference



from datetime import datetime
from typing import Any, Callable, Dict, Literal, Optional, Union

from supervisely.app import StateJson
from supervisely.app.widgets import Widget


class DateTimePicker(Widget):
    """DateTimePicker is a widget in Supervisely that allows you to choose a date and time on the UI.

    Read about it in `Developer Portal <https://developer.supervisely.com/app-development/widgets/input/datetimepicker>`_
        (including screenshots and examples).

    :param value: initial value
    :type value: Optional[Union[int, str, datetime, list, tuple]]
    :param readonly: if True, date picker will be readonly
    :type readonly: Optional[bool]
    :param disabled: if True, date picker will be disabled
    :type disabled: Optional[bool]
    :param editable: if True, date picker will be editable
    :type editable: Optional[bool]
    :param clearable: if True, date picker will be clearable
    :type clearable: Optional[bool]
    :param size: size of the input box, one of: large, small, mini
    :type size: Optional[Literal["large", "small", "mini"]]
    :param placeholder: placeholder text, default: "Select date and time"
    :type placeholder: Optional[str]
    :param w_type: picker type, one of: year, month, date, datetime, week, datetimerange, daterange, default: datetime
    :type w_type: Optional[Literal["year", "month", "date", "datetime", "week", "datetimerange", "daterange"]]
    :param format: date format, one of: yyyy, MM, dd, HH, mm, ss, default: yyyy-MM-dd HH:mm:ss
    :type format: Optional[Literal["yyyy", "MM", "dd", "HH", "mm", "ss"]]
    :param widget_id: An identifier of the widget.
    :type widget_id: str, optional

    :Usage example:
    .. code-block:: python

            from supervisely.app.widgets import DateTimePicker

            date_time_picker = DateTimePicker(
                value=datetime.now(),
                readonly=False,
                disabled=False,
                editable=False,
                clearable=True,
                size="small",
                placeholder="Select date and time",
                w_type="datetime",
                format="yyyy-MM-dd HH:mm:ss",
            )
    """

    class Routes:
        VALUE_CHANGED = "value_changed"

    def __init__(
        self,
        value: Optional[Union[int, str, list, tuple]] = None,
        readonly: Optional[bool] = False,
        disabled: Optional[bool] = False,
        editable: Optional[bool] = False,
        clearable: Optional[bool] = True,
        size: Optional[Literal["large", "small", "mini"]] = None,
        placeholder: Optional[str] = "Select date and time",
        w_type: Optional[
            Literal["year", "month", "date", "datetime", "week", "datetimerange", "daterange"]
        ] = "datetime",
        format: Optional[Literal["yyyy", "MM", "dd", "HH", "mm", "ss"]] = "yyyy-MM-dd HH:mm:ss",
        widget_id: Optional[str] = None,
    ):
        self._value = value
        self._readonly = readonly
        self._disabled = disabled
        self._editable = editable
        self._clearable = clearable
        self._size = size
        self._placeholder = placeholder
        self._w_type = w_type
        self._format = format
        self._changes_handled = False

        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self) -> Dict[str, Union[str, bool]]:
        """Returns dictionary with widget data, which defines the appearance and behavior of the widget.

        Dictionary contains the following fields:
            - placeholder: placeholder text
            - size: size of the input box, one of: large, small, mini
            - readonly: if True, date picker will be readonly
            - disabled: if True, date picker will be disabled
            - editable: if True, date picker will be editable
            - clearable: if True, date picker will be clearable
            - type: picker type, one of: year, month, date, datetime, week, datetimerange, daterange
            - format: date format, one of: yyyy, MM, dd, HH, mm, ss

        :return: dictionary with widget data
        :rtype: Dict[str, Union[str, bool]]
        """
        return {
            "placeholder": self._placeholder,
            "size": self._size,
            "readonly": self._readonly,
            "disabled": self._disabled,
            "editable": self._editable,
            "clearable": self._clearable,
            "type": self._w_type,
            "format": self._format,
        }

    def get_json_state(self) -> Dict[str, Union[str, list]]:
        """Returns dictionary with widget state.

        Dictionary contains the following fields:
            - value: current value

        :return: dictionary with widget state
        :rtype: Dict[str, Union[str, list]]
        """
        return {"value": self._value}

    def get_value(self) -> Union[int, str, datetime, list, tuple]:
        """Returns current value.

        :return: current value
        :rtype: Union[int, str, datetime, list, tuple]
        """
        if "value" not in StateJson()[self.widget_id].keys():
            return None
        value = StateJson()[self.widget_id]["value"]
        if self._w_type in ["datetimerange", "daterange"] and any(
            [bool(date) is False for date in value]
        ):
            return None
        elif self._w_type not in ["datetimerange", "daterange"] and value == "":
            return None
        return value

    def set_value(self, value: Union[int, str, datetime, list, tuple]) -> None:
        """Sets current value.

        :param value: current value
        :type value: Union[int, str, datetime, list, tuple]
        :raises ValueError: if value type is not supported
        :raises ValueError: if value length is not equal 2
        """
        if self._w_type in ["year", "month", "date", "datetime", "week"]:
            if type(value) not in [int, str, datetime]:
                raise ValueError(
                    f'Datetime picker type "{self._w_type}" does not support value "{value}" of type: '
                    f'"{str(type(value))}". Value type has to be one of: ["int", "str", "datetime].'
                )
            if isinstance(value, datetime):
                value = str(value)

        if self._w_type in ["datetimerange", "daterange"]:
            if type(value) not in [list, tuple]:
                raise ValueError(
                    f'Datetime picker type "{self._w_type}" does not support value "{value}" of type: '
                    f'"{str(type(value))}". Value type has to be one of: ["list", "tuple"].'
                )
            if len(value) != 2:
                raise ValueError(f"Value length has to be equal 2: {len(value)} != 2")
            value = [str(val) if isinstance(val, datetime) else val for val in value]

        self._value = value
        StateJson()[self.widget_id]["value"] = self._value
        StateJson().send_changes()

    def value_changed(
        self, func: Callable[[Union[int, str, datetime, list, tuple]], Any]
    ) -> Callable[[], None]:
        """Decorator for the function that will be called when the value of the widget is changed.

        :param func: function that will be called when the value of the widget is changed
        :type func: Callable[[Union[int, str, datetime, list, tuple]], Any]
        :return: decorated function
        :rtype: Callable[[], None]
        """
        route_path = self.get_route_path(DateTimePicker.Routes.VALUE_CHANGED)
        server = self._sly_app.get_server()
        self._changes_handled = True

        @server.post(route_path)
        def _click():
            res = self.get_value()
            self._value = res
            func(res)

        return _click



import functools
import json
import time
from pathlib import Path
from queue import Queue
from threading import Event, Thread
from typing import Any, Dict, List, Optional, Union

import numpy as np
from fastapi import BackgroundTasks, Form, Request, UploadFile

import supervisely as sly
import supervisely.nn.inference.tracking.functional as F
from supervisely.annotation.label import Geometry, Label
from supervisely.nn.inference import Inference
from supervisely.nn.inference.tracking.tracker_interface import TrackerInterface
from supervisely.nn.prediction_dto import Prediction, PredictionBBox


class BBoxTracking(Inference):
    def __init__(
        self,
        model_dir: Optional[str] = None,
        custom_inference_settings: Optional[Union[Dict[str, Any], str]] = None,
    ):
        Inference.__init__(
            self,
            model_dir,
            custom_inference_settings,
            sliding_window_mode=None,
            use_gui=False,
        )

        try:
            self.load_on_device(model_dir, "cuda")
        except RuntimeError:
            self.load_on_device(model_dir, "cpu")
            sly.logger.warn("Failed to load model on CUDA device.")

        sly.logger.debug(
            "Smart cache params",
            extra={"ttl": sly.env.smart_cache_ttl(), "maxsize": sly.env.smart_cache_size()},
        )

    def get_info(self):
        info = super().get_info()
        info["task type"] = "tracking"
        return info

    def _deserialize_geometry(self, data: dict):
        geometry_type_str = data["type"]
        geometry_json = data["data"]
        return sly.deserialize_geometry(geometry_type_str, geometry_json)

    def _track(self, api: sly.Api, context: dict, notify_annotation_tool: bool):
        video_interface = TrackerInterface(
            context=context,
            api=api,
            load_all_frames=False,
            frame_loader=self.cache.download_frame,
            frames_loader=self.cache.download_frames,
            should_notify=notify_annotation_tool,
        )

        range_of_frames = [
            video_interface.frames_indexes[0],
            video_interface.frames_indexes[-1],
        ]

        if self.cache.is_persistent:
            self.cache.run_cache_task_manually(
                api,
                None,
                video_id=video_interface.video_id,
            )
        else:
            # if cache is not persistent, run cache task for range of frames
            self.cache.run_cache_task_manually(
                api,
                [range_of_frames],
                video_id=video_interface.video_id,
            )

        api.logger.info("Start tracking.")

        def _upload_loop(q: Queue, stop_event: Event, video_interface: TrackerInterface):
            try:
                while True:
                    items = []
                    while not q.empty():
                        items.append(q.get_nowait())
                    if len(items) > 0:
                        video_interface.add_object_geometries_on_frames(*list(zip(*items)))
                    if stop_event.is_set():
                        video_interface._notify(True, task="stop tracking")
                        return
                    time.sleep(1)
            except Exception as e:
                api.logger.error("Error in upload loop: %s", str(e), exc_info=True)
                video_interface._notify(True, task="stop tracking")
                video_interface.global_stop_indicatior = True
                raise

        upload_queue = Queue()
        stop_upload_event = Event()
        Thread(
            target=_upload_loop,
            args=[upload_queue, stop_upload_event, video_interface],
            daemon=True,
        ).start()

        try:
            for fig_id, obj_id in zip(
                video_interface.geometries.keys(),
                video_interface.object_ids,
            ):
                init = False
                for _ in video_interface.frames_loader_generator():
                    geom = video_interface.geometries[fig_id]
                    if not isinstance(geom, sly.Rectangle):
                        stop_upload_event.set()
                        raise TypeError(f"Tracking does not work with {geom.geometry_name()}.")

                    imgs = video_interface.frames
                    target = PredictionBBox(
                        "",  # TODO: can this be useful?
                        [geom.top, geom.left, geom.bottom, geom.right],
                        None,
                    )

                    if not init:
                        self.initialize(imgs[0], target)
                        init = True

                    geometry = self.predict(
                        rgb_image=imgs[-1],
                        prev_rgb_image=imgs[0],
                        target_bbox=target,
                        settings=self.custom_inference_settings_dict,
                    )
                    sly_geometry = self._to_sly_geometry(geometry)
                    upload_queue.put(
                        (sly_geometry, obj_id, video_interface._cur_frames_indexes[-1])
                    )

                    if video_interface.global_stop_indicatior:
                        stop_upload_event.set()
                        return

                api.logger.info(f"Figure #{fig_id} tracked.")
        except Exception:
            stop_upload_event.set()
            raise
        stop_upload_event.set()

    def _track_api(self, api: sly.Api, context: dict):
        # unused fields:
        context["trackId"] = "auto"
        context["objectIds"] = []
        context["figureIds"] = []
        if "direction" not in context:
            context["direction"] = "forward"

        input_bboxes: list = context["input_geometries"]

        video_interface = TrackerInterface(
            context=context,
            api=api,
            load_all_frames=False,
            frame_loader=self.cache.download_frame,
            frames_loader=self.cache.download_frames,
            should_notify=False,
        )

        range_of_frames = [
            video_interface.frames_indexes[0],
            video_interface.frames_indexes[-1],
        ]

        if self.cache.is_persistent:
            # if cache is persistent, run cache task for whole video
            self.cache.run_cache_task_manually(
                api,
                None,
                video_id=video_interface.video_id,
            )
        else:
            # if cache is not persistent, run cache task for range of frames
            self.cache.run_cache_task_manually(
                api,
                [range_of_frames],
                video_id=video_interface.video_id,
            )

        api.logger.info("Start tracking.")

        predictions = []
        for input_geom in input_bboxes:
            input_bbox = input_geom["data"]
            bbox = sly.Rectangle.from_json(input_bbox)
            predictions_for_object = []
            init = False
            for _ in video_interface.frames_loader_generator():
                imgs = video_interface.frames
                target = PredictionBBox(
                    "",  # TODO: can this be useful?
                    [bbox.top, bbox.left, bbox.bottom, bbox.right],
                    None,
                )

                if not init:
                    self.initialize(imgs[0], target)
                    init = True

                geometry = self.predict(
                    rgb_image=imgs[-1],
                    prev_rgb_image=imgs[0],
                    target_bbox=target,
                    settings=self.custom_inference_settings_dict,
                )
                sly_geometry = self._to_sly_geometry(geometry)

                predictions_for_object.append(
                    {"type": sly_geometry.geometry_name(), "data": sly_geometry.to_json()}
                )
            predictions.append(predictions_for_object)

        # predictions must be NxK bboxes: N=number of frames, K=number of objects
        predictions = list(map(list, zip(*predictions)))
        return predictions

    def _inference(self, frames: List[np.ndarray], geometries: List[Geometry], settings: dict):
        updated_settings = {
            **self.custom_inference_settings_dict,
            **settings,
        }
        results = [[] for _ in range(len(frames) - 1)]
        for geometry in geometries:
            if not isinstance(geometry, sly.Rectangle):
                raise TypeError(f"Tracking does not work with {geometry.geometry_name()}.")
            target = PredictionBBox(
                "",
                [geometry.top, geometry.left, geometry.bottom, geometry.right],
                None,
            )
            self.initialize(frames[0], target)
            for i in range(len(frames) - 1):
                pred_geometry = self.predict(
                    rgb_image=frames[i + 1],
                    prev_rgb_image=frames[i],
                    target_bbox=target,
                    settings=updated_settings,
                )
                sly_pred_geometry = self._to_sly_geometry(pred_geometry)
                results[i].append(
                    {"type": sly.Rectangle.geometry_name(), "data": sly_pred_geometry.to_json()}
                )
        return results

    def _track_api_files(
        self, request: Request, files: List[UploadFile], settings: str = Form("{}")
    ):
        state = json.loads(settings)
        sly.logger.info(f"Start tracking with settings: {state}.")
        video_id = state["video_id"]
        frame_indexes = list(
            range(state["frame_index"], state["frame_index"] + state["frames"] + 1)
        )
        geometries = map(self._deserialize_geometry, state["input_geometries"])
        frames = []
        for file, frame_idx in zip(files, frame_indexes):
            img_bytes = file.file.read()
            frame = sly.image.read_bytes(img_bytes)
            frames.append(frame)
        sly.logger.info("Start tracking.")
        return self._inference(frames, geometries, state)

    def serve(self):
        super().serve()
        server = self._app.get_server()
        self.cache.add_cache_endpoint(server)
        self.cache.add_cache_files_endpoint(server)

        def send_error_data(func):
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                value = None
                try:
                    value = func(*args, **kwargs)
                except Exception as exc:
                    request: Request = args[0]
                    context = request.state.context
                    api: sly.Api = request.state.api
                    track_id = context["trackId"]
                    api.logger.error(f"An error occured: {repr(exc)}")

                    api.post(
                        "videos.notify-annotation-tool",
                        data={
                            "type": "videos:tracking-error",
                            "data": {
                                "trackId": track_id,
                                "error": {"message": repr(exc)},
                            },
                        },
                    )
                return value

            return wrapper

        @send_error_data
        def track(request: Request):
            return self._track(
                request.state.api, request.state.context, notify_annotation_tool=True
            )

        @server.post("/track")
        def start_track(request: Request, task: BackgroundTasks):
            task.add_task(track, request)
            return {"message": "Track task started."}

        @server.post("/track-api")
        def track_api(request: Request):
            sly.logger.info("Start tracking.")
            return self._track_api(request.state.api, request.state.context)

        @server.post("/track-api-files")
        def track_api_files(
            request: Request,
            files: List[UploadFile],
            settings: str = Form("{}"),
        ):
            return self._track_api_files(request, files, settings)

    def initialize(self, init_rgb_image: np.ndarray, target_bbox: PredictionBBox) -> None:
        """
        Initializing the tracker with a new object.

        :param init_rgb_image: frame with object
        :type init_rgb_image: np.ndarray
        :param target_bbox: initial bbox
        :type target_bbox: PredictionBBox
        """
        raise NotImplementedError

    def predict(
        self,
        rgb_image: np.ndarray,
        settings: Dict[str, Any],
        prev_rgb_image: np.ndarray,
        target_bbox: PredictionBBox,
    ) -> PredictionBBox:
        """
        SOT prediction

        :param rgb_image: search frame
        :type rgb_image: np.ndarray
        :param settings: model parameters
        :type settings: Dict[str, Any]
        :param init_rgb_image: previous frame with object
        :type init_rgb_image: np.ndarray
        :param target_bbox: bbox added on previous step
        :type target_bbox: PredictionBBox
        :return: predicted annotation
        :rtype: PredictionBBox
        """
        raise NotImplementedError

    def visualize(
        self,
        predictions: List[PredictionBBox],
        images: List[np.ndarray],
        vis_path: str,
        thickness: int = 2,
        classes_whitelist: Optional[List[str]] = None,
    ):
        vis_path = Path(vis_path)

        for i, (pred, image) in enumerate(zip(predictions, images)):
            out_path = vis_path / f"img_{i}.jpg"
            ann = self._predictions_to_annotation(image, [pred], classes_whitelist)
            ann.draw_pretty(
                bitmap=image,
                color=(255, 0, 0),
                thickness=thickness,
                output_path=str(out_path),
                fill_rectangles=False,
            )

    def _to_sly_geometry(self, dto: PredictionBBox) -> sly.Rectangle:
        top, left, bottom, right = dto.bbox_tlbr
        geometry = sly.Rectangle(top=top, left=left, bottom=bottom, right=right)
        return geometry

    def _create_label(self, dto: PredictionBBox) -> sly.Rectangle:
        geometry = self._to_sly_geometry(dto)
        return Label(geometry, sly.ObjClass("", sly.Rectangle))

    def _get_obj_class_shape(self):
        return sly.Rectangle

    def _predictions_to_annotation(
        self,
        image: np.ndarray,
        predictions: List[Prediction],
        classes_whitelist: Optional[List[str]] = None,
    ) -> sly.Annotation:
        labels = []
        for prediction in predictions:
            if (
                not classes_whitelist in (None, "all")
                and prediction.class_name not in classes_whitelist
            ):
                continue
            label = self._create_label(prediction)
            if label is None:
                # for example empty mask
                continue
            if isinstance(label, list):
                labels.extend(label)
                continue
            labels.append(label)

        # create annotation with correct image resolution
        ann = sly.Annotation(img_size=image.shape[:2])
        ann = ann.add_labels(labels)
        return ann



import functools
import json
import time
from pathlib import Path
from queue import Queue
from threading import Event, Thread
from typing import Any, Dict, List, Optional, Union

import numpy as np
from fastapi import BackgroundTasks, Form, Request, UploadFile

import supervisely as sly
import supervisely.nn.inference.tracking.functional as F
from supervisely.annotation.label import Geometry, Label
from supervisely.nn.inference import Inference
from supervisely.nn.inference.tracking.tracker_interface import TrackerInterface
from supervisely.nn.prediction_dto import Prediction, PredictionPoint


class PointTracking(Inference):
    def __init__(
        self,
        model_dir: Optional[str] = None,
        custom_inference_settings: Optional[Union[Dict[str, Any], str]] = None,
    ):
        Inference.__init__(
            self,
            model_dir,
            custom_inference_settings,
            sliding_window_mode=None,
            use_gui=False,
        )

        try:
            self.load_on_device(model_dir, "cuda")
        except RuntimeError:
            self.load_on_device(model_dir, "cpu")
            sly.logger.warn("Failed to load model on CUDA device.")

        sly.logger.debug(
            "Smart cache params",
            extra={
                "ttl": sly.env.smart_cache_ttl(),
                "maxsize": sly.env.smart_cache_size(),
                "path": sly.env.smart_cache_container_dir(),
            },
        )

    def _deserialize_geometry(self, data: dict):
        geometry_type_str = data["type"]
        geometry_json = data["data"]
        return sly.deserialize_geometry(geometry_type_str, geometry_json)

    def get_info(self):
        info = super().get_info()
        info["task type"] = "tracking"
        # recommended parameters:
        # info["model_name"] = ""
        # info["checkpoint_name"] = ""
        # info["pretrained_on_dataset"] = ""
        # info["device"] = ""
        return info

    def track_api(self, api: sly.Api, context: dict):
        # unused fields:
        context["trackId"] = "auto"
        context["objectIds"] = []
        context["figureIds"] = []
        if "direction" not in context:
            context["direction"] = "forward"

        input_geometries: list = context["input_geometries"]

        if self.custom_inference_settings_dict.get("load_all_frames"):
            load_all_frames = True
        else:
            load_all_frames = False
        video_interface = TrackerInterface(
            context=context,
            api=api,
            load_all_frames=load_all_frames,
            frame_loader=self.cache.download_frame,
            frames_loader=self.cache.download_frames,
            should_notify=False,
        )

        range_of_frames = [
            video_interface.frames_indexes[0],
            video_interface.frames_indexes[-1],
        ]

        if self.cache.is_persistent:
            # if cache is persistent, run cache task for whole video
            self.cache.run_cache_task_manually(
                api,
                None,
                video_id=video_interface.video_id,
            )
        else:
            # if cache is not persistent, run cache task for range of frames
            self.cache.run_cache_task_manually(
                api,
                [range_of_frames],
                video_id=video_interface.video_id,
            )

        api.logger.info("Start tracking.")

        predictions = []
        for _ in video_interface.frames_loader_generator():
            for input_geom in input_geometries:
                geom = self._deserialize_geometry(input_geom)
                if isinstance(geom, sly.Point):
                    geometries = self._predict_point_geometries(
                        geom,
                        video_interface.frames,
                    )
                elif isinstance(geom, sly.Polygon):
                    if len(geom.interior) > 0:
                        raise ValueError("Can't track polygons with interior.")
                    geometries = self._predict_polygon_geometries(
                        geom,
                        video_interface.frames,
                    )
                elif isinstance(geom, sly.GraphNodes):
                    geometries = self._predict_graph_geometries(
                        geom,
                        video_interface.frames,
                    )
                elif isinstance(geom, sly.Polyline):
                    geometries = self._predict_polyline_geometries(
                        geom,
                        video_interface.frames,
                    )
                else:
                    raise TypeError(f"Tracking does not work with {geom.geometry_name()}.")

                if video_interface.global_stop_indicatior:
                    return

                geometries = [{"type": g.geometry_name(), "data": g.to_json()} for g in geometries]
                predictions.append(geometries)

        # predictions must be NxK figures: N=number of frames, K=number of objects
        predictions = list(map(list, zip(*predictions)))
        return predictions

    def _inference(self, frames: List[np.ndarray], geometries: List[Geometry], settings: dict):
        updated_settings = {
            **self.custom_inference_settings_dict,
            **settings,
        }
        results = [[] for _ in range(len(frames) - 1)]
        for geometry in geometries:
            if isinstance(geometry, sly.Point):
                predictions = self._predict_point_geometries(geometry, frames, updated_settings)
            elif isinstance(geometry, sly.Polygon):
                if len(geometry.interior) > 0:
                    raise ValueError("Can't track polygons with interior.")
                predictions = self._predict_polygon_geometries(
                    geometry,
                    frames,
                    updated_settings,
                )
            elif isinstance(geometry, sly.GraphNodes):
                predictions = self._predict_graph_geometries(
                    geometry,
                    frames,
                    updated_settings,
                )
            elif isinstance(geometry, sly.Polyline):
                predictions = self._predict_polyline_geometries(
                    geometry,
                    frames,
                    updated_settings,
                )
            else:
                raise TypeError(f"Tracking does not work with {geometry.geometry_name()}.")

            for i, prediction in enumerate(predictions):
                results[i].append({"type": geometry.geometry_name(), "data": prediction.to_json()})

        return results

    def track_api_cached(self, request: Request, context: dict):
        sly.logger.info(f"Start tracking with settings: {context}.")
        video_id = context["video_id"]
        frame_indexes = list(
            range(context["frame_index"], context["frame_index"] + context["frames"] + 1)
        )
        geometries = map(self._deserialize_geometry, context["input_geometries"])
        frames = self.cache.get_frames_from_cache(video_id, frame_indexes)
        return self._inference(frames, geometries, context)

    def _track_api_files(
        self, request: Request, files: List[UploadFile], settings: str = Form("{}")
    ):
        state = json.loads(settings)
        sly.logger.info(f"Start tracking with settings: {state}.")
        video_id = state["video_id"]
        frame_indexes = list(
            range(state["frame_index"], state["frame_index"] + state["frames"] + 1)
        )
        geometries = map(self._deserialize_geometry, state["input_geometries"])
        frames = []
        for file, frame_idx in zip(files, frame_indexes):
            img_bytes = file.file.read()
            frame = sly.image.read_bytes(img_bytes)
            frames.append(frame)
        sly.logger.info("Start tracking.")
        return self._inference(frames, geometries, state)

    def serve(self):
        super().serve()
        server = self._app.get_server()
        self.cache.add_cache_endpoint(server)
        self.cache.add_cache_files_endpoint(server)

        @server.post("/track")
        def start_track(request: Request, task: BackgroundTasks):
            task.add_task(track, request)
            return {"message": "Track task started."}

        @server.post("/track-api")
        def track_api(request: Request):
            return self.track_api(request.state.api, request.state.context)

        @server.post("/track-api-files")
        def track_api_frames_files(
            request: Request,
            files: List[UploadFile],
            settings: str = Form("{}"),
        ):
            return self._track_api_files(request, files, settings)

        def send_error_data(func):
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                value = None
                try:
                    value = func(*args, **kwargs)
                except Exception as exc:
                    request: Request = args[0]
                    context = request.state.context
                    api: sly.Api = request.state.api
                    track_id = context["trackId"]
                    api.logger.error(f"An error occured: {repr(exc)}")

                    api.post(
                        "videos.notify-annotation-tool",
                        data={
                            "type": "videos:tracking-error",
                            "data": {
                                "trackId": track_id,
                                "error": {"message": repr(exc)},
                            },
                        },
                    )
                return value

            return wrapper

        @send_error_data
        def track(request: Request):
            context = request.state.context
            api: sly.Api = request.state.api

            if self.custom_inference_settings_dict.get("load_all_frames"):
                load_all_frames = True
            else:
                load_all_frames = False
            video_interface = TrackerInterface(
                context=context,
                api=api,
                load_all_frames=load_all_frames,
                frame_loader=self.cache.download_frame,
                frames_loader=self.cache.download_frames,
            )

            range_of_frames = [
                video_interface.frames_indexes[0],
                video_interface.frames_indexes[-1],
            ]

            self.cache.run_cache_task_manually(
                api,
                [range_of_frames],
                video_id=video_interface.video_id,
            )
            api.logger.info("Start tracking.")

            def _upload_loop(q: Queue, stop_event: Event, video_interface: TrackerInterface):
                try:
                    while True:
                        items = []
                        while not q.empty():
                            items.append(q.get_nowait())
                        if len(items) > 0:
                            video_interface.add_object_geometries_on_frames(*list(zip(*items)))
                        if stop_event.is_set():
                            video_interface._notify(True, task="stop tracking")
                            return
                        time.sleep(1)
                except Exception as e:
                    api.logger.error("Error in upload loop: %s", str(e), exc_info=True)
                    video_interface._notify(True, task="stop tracking")
                    video_interface.global_stop_indicatior = True
                    raise

            upload_queue = Queue()
            stop_upload_event = Event()
            Thread(
                target=_upload_loop,
                args=[upload_queue, stop_upload_event, video_interface],
                daemon=True,
            ).start()
            try:
                for _ in video_interface.frames_loader_generator():
                    for (fig_id, geom), obj_id in zip(
                        video_interface.geometries.items(),
                        video_interface.object_ids,
                    ):
                        if isinstance(geom, sly.Point):
                            geometries = self._predict_point_geometries(
                                geom,
                                video_interface.frames_with_notification,
                            )
                        elif isinstance(geom, sly.Polygon):
                            if len(geom.interior) > 0:
                                stop_upload_event.set()
                                raise ValueError("Can't track polygons with interior.")
                            geometries = self._predict_polygon_geometries(
                                geom,
                                video_interface.frames_with_notification,
                            )
                        elif isinstance(geom, sly.GraphNodes):
                            geometries = self._predict_graph_geometries(
                                geom,
                                video_interface.frames_with_notification,
                            )
                        elif isinstance(geom, sly.Polyline):
                            geometries = self._predict_polyline_geometries(
                                geom,
                                video_interface.frames_with_notification,
                            )
                        else:
                            raise TypeError(f"Tracking does not work with {geom.geometry_name()}.")

                        for frame_idx, geometry in zip(
                            video_interface._cur_frames_indexes[1:], geometries
                        ):
                            upload_queue.put(
                                (
                                    geometry,
                                    obj_id,
                                    frame_idx,
                                )
                            )
                        api.logger.info(f"Object #{obj_id} tracked.")

                        if video_interface.global_stop_indicatior:
                            stop_upload_event.set()
                            return
            except Exception:
                stop_upload_event.set()
                raise
            stop_upload_event.set()

    def predict(
        self,
        rgb_images: List[np.ndarray],
        settings: Dict[str, Any],
        start_object: Union[PredictionPoint, List[PredictionPoint]],
    ) -> List[PredictionPoint]:
        """
        Track point on given frames.

        :param rgb_images: RGB frames, `m` frames
        :type rgb_images: List[np.array]
        :param settings: model parameters
        :type settings: Dict[str, Any]
        :param start_object: point to track on the initial frame
        :type start_object: PredictionPoint
        :return: predicted points for frame range (0, m]; `m-1` prediction in total
        :rtype: List[PredictionPoint]
        """
        raise NotImplementedError

    def predict_batch(
        self,
        rgb_images: List[np.ndarray],
        settings: Dict[str, Any],
        start_objects: List[PredictionPoint],
    ) -> List[List[PredictionPoint]]:
        """
        Track points on given frames.

        :param rgb_images: RGB frames, `m` frames
        :type rgb_images: List[np.array]
        :param settings: model parameters
        :type settings: Dict[str, Any]
        :param start_objects: points to track on the initial frame
        :type start_objects: List[PredictionPoint]
        :return: predicted points for frame range (0, m]; `m-1` prediction in total
        :rtype: List[List[PredictionPoint]]
        """
        raise NotImplementedError

    def visualize(
        self,
        predictions: List[PredictionPoint],
        images: List[np.ndarray],
        vis_path: str,
        thickness: int = 2,
        classes_whitelist: List[str] = None,
    ):
        vis_path = Path(vis_path)

        for i, (pred, image) in enumerate(zip(predictions, images)):
            out_path = vis_path / f"img_{i}.jpg"
            ann = self._predictions_to_annotation(image, [pred], classes_whitelist)
            ann.draw_pretty(
                bitmap=image,
                color=(255, 0, 0),
                thickness=thickness,
                output_path=str(out_path),
                fill_rectangles=False,
            )

    def _create_label(self, dto: PredictionPoint) -> sly.Point:
        geometry = sly.Point(row=dto.row, col=dto.col)
        return Label(geometry, sly.ObjClass("", sly.Point))

    def _get_obj_class_shape(self):
        return sly.Point

    def _predict_point_geometries(
        self,
        geom: sly.Point,
        frames: List[np.ndarray],
        settings: Dict[str, Any] = None,
    ) -> List[sly.Point]:
        if settings is None:
            settings = self.custom_inference_settings_dict
        pp_geom = PredictionPoint("point", col=geom.col, row=geom.row)
        if type(self).predict_batch == PointTracking.predict_batch:
            # if predict_batch is not implemented, we can't use it
            predicted = self.predict(
                frames,
                settings,
                pp_geom,
            )
        else:
            predicted = self.predict_batch(
                frames,
                settings,
                pp_geom,
            )
            predicted = [pred[0] for pred in predicted]
        return F.dto_points_to_sly_points(predicted)

    def _predict_polygon_geometries(
        self,
        geom: sly.Polygon,
        frames: List[np.ndarray],
        settings: Dict[str, Any] = None,
    ) -> List[sly.Polygon]:
        if settings is None:
            settings = self.custom_inference_settings_dict
        polygon_points = F.numpy_to_dto_point(geom.exterior_np, "polygon")

        if type(self).predict_batch == PointTracking.predict_batch:
            # if predict_batch is not implemented, we can't use it
            points = list(
                zip(
                    *[
                        self.predict(frames, settings, polygon_point)
                        for polygon_point in polygon_points
                    ]
                )
            )
        else:
            points: List[List[Prediction]] = self.predict_batch(
                frames,
                settings,
                polygon_points,
            )
        points_loc = [F.dto_points_to_point_location(frame_points) for frame_points in points]
        return F.exteriors_to_sly_polygons(points_loc)

    def _predict_graph_geometries(
        self,
        geom: sly.GraphNodes,
        frames: List[np.ndarray],
        settings: Dict[str, Any] = None,
    ) -> List[sly.GraphNodes]:
        if settings is None:
            settings = self.custom_inference_settings_dict
        points, pids = F.graph_to_dto_points(geom)

        if type(self).predict_batch == PointTracking.predict_batch:
            # if predict_batch is not implemented, we can't use it
            preds = list(zip(*[self.predict(frames, settings, point) for point in points]))
        else:
            preds: List[List[PredictionPoint]] = self.predict_batch(
                frames,
                settings,
                points,
            )

        nodes = []
        for frame_preds in preds:
            frame_nodes = []
            for pred, pid in zip(frame_preds, pids):
                frame_nodes.extend(F.dto_points_to_sly_nodes([pred], pid))
            nodes.append(frame_nodes)

        return F.nodes_to_sly_graph(nodes)

    def _predict_polyline_geometries(
        self,
        geom: sly.Polyline,
        frames: List[np.ndarray],
        settings: Dict[str, Any] = None,
    ) -> List[sly.Polyline]:
        if settings is None:
            settings = self.custom_inference_settings_dict
        polyline_points = F.numpy_to_dto_point(geom.exterior_np, "polyline")
        if type(self).predict_batch == PointTracking.predict_batch:
            # if predict_batch is not implemented, we can't use it
            preds = list(
                zip(
                    *[
                        self.predict(frames, settings, polyline_point)
                        for polyline_point in polyline_points
                    ]
                )
            )
        else:
            preds = self.predict_batch(
                frames,
                settings,
                polyline_points,
            )
        points_loc = [F.dto_points_to_point_location(frame_points) for frame_points in preds]
        return F.exterior_to_sly_polyline(points_loc)

    def _predictions_to_annotation(
        self, image: np.ndarray, predictions: List[Prediction], classes_whitelist: List[str] = None
    ) -> sly.Annotation:
        labels = []
        for prediction in predictions:
            if (
                not classes_whitelist in (None, "all")
                and prediction.class_name not in classes_whitelist
            ):
                continue
            label = self._create_label(prediction)
            if label is None:
                # for example empty mask
                continue
            if isinstance(label, list):
                labels.extend(label)
                continue
            labels.append(label)

        # create annotation with correct image resolution
        ann = sly.Annotation(img_size=image.shape[:2])
        ann = ann.add_labels(labels)
        return ann



from supervisely.app import DataJson, StateJson
from supervisely.app.widgets import Widget
from typing import List, Union

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


class Pagination(Widget):
    class Routes:
        CURRENT_CHANGE = "current_change"
        SIZE_CHANGE = "size_change"

    def __init__(
        self,
        total: int,
        page_size: int = 10,
        current_page: int = 1,
        layout: Union[
            str, List[Literal["sizes", "prev", "pager", "next", "jumper", "->", "total"]]
        ] = "prev, pager, next",
        compact: bool = False,
        page_size_options: List[int] = [10, 20, 30, 40, 50, 100],
        widget_id: str = None,
    ):
        self._total = total
        self._page_size = page_size
        self._current_page = current_page
        self._layout = ", ".join(layout) if type(layout) == list else layout
        self._compact = compact
        self._page_size_options = page_size_options

        self._current_change = False
        self._size_change = False
        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self):
        return {
            "layout": self._layout,
            "total": self._total,
            "compact": self._compact,
            "pageSizeOptions": self._page_size_options,
        }

    def get_json_state(self):
        return {
            "currentPage": self._current_page,
            "pageSize": self._page_size,
        }

    def get_current_page(self):
        return StateJson()[self.widget_id]["currentPage"]

    def set_current_page(self, page: int):
        max_page = self.get_total() // self.get_page_size() + 1
        if page > max_page:
            page = max_page
        if page < 1:
            page = 1
        if type(page) != int:
            raise TypeError("Current page value must be int")
        self._current_page = page
        StateJson()[self.widget_id]["currentPage"] = self._current_page
        StateJson().send_changes()

    def get_total(self):
        return DataJson()[self.widget_id]["total"]

    def set_total(self, value: int):
        if type(value) != int:
            raise TypeError("Total value must be int")
        self._total = value
        DataJson()[self.widget_id]["total"] = self._total
        DataJson().send_changes()

    def get_page_size(self):
        return StateJson()[self.widget_id]["pageSize"]

    def set_page_size(self, value: int):
        if type(value) != int:
            raise TypeError("Page size value must be int")

        if value not in self.get_page_size_options():
            raise ValueError("Page size value must be in page size options")

        self._current_page = value
        StateJson()[self.widget_id]["pageSize"] = self._current_page
        StateJson().send_changes()

    def get_page_size_options(self):
        return DataJson()[self.widget_id]["pageSizeOptions"]

    def set_page_size_options(self, value: List[int]):
        if type(value) != list or not all(type(x) == int for x in value):
            raise TypeError("Page size options value must be list of integers")
        self._page_size_options = value
        DataJson()[self.widget_id]["pageSizeOptions"] = self._page_size_options
        DataJson().send_changes()

    def set_layout(
        self,
        value: Union[
            str, List[Literal["sizes", "prev", "pager", "next", "jumper", "->", "total", "slot"]]
        ],
    ):
        if type(value) == list:
            value = ", ".join(value)
        if type(value) != str:
            raise TypeError("Layout value must be str or list")
        self._layout = value
        DataJson()[self.widget_id]["layout"] = self._layout
        DataJson().send_changes()

    def page_changed(self, func):
        route_path = self.get_route_path(Pagination.Routes.CURRENT_CHANGE)
        server = self._sly_app.get_server()
        self._current_change = True

        @server.post(route_path)
        def _click():
            res = self.get_current_page()
            func(res)

        return _click

    def page_size_changed(self, func):
        route_path = self.get_route_path(Pagination.Routes.SIZE_CHANGE)
        server = self._sly_app.get_server()
        self._size_change = True

        @server.post(route_path)
        def _click():
            res = self.get_page_size()
            func(res)

        return _click



import os
import re

import requests
from pkg_resources import DistributionNotFound, get_distribution
from setuptools import find_packages, setup

# @TODO: change manifest location


def read(fname):
    with open(os.path.join(os.path.dirname(__file__), fname)) as fin:
        return fin.read()


response = requests.get("https://api.github.com/repos/supervisely/supervisely/releases/latest")
version = response.json()["tag_name"]


INSTALL_REQUIRES = [
    "cachetools<5.0.0",
    "numpy>=1.19, <2.0.0",
    "opencv-python>=4.5.5.62, <5.0.0.0",
    "PTable>=0.9.2, <1.0.0",
    "pillow>=5.4.1, <=10.2.0",
    "protobuf>=3.14.0, <=3.20.3",
    "python-json-logger>=0.1.11, <3.0.0",
    "requests>=2.27.1, <3.0.0",
    "requests-toolbelt>=0.9.1",  # , <1.0.0
    "Shapely>=1.7.1, <=2.0.2",
    "bidict>=0.21.2, <1.0.0",
    "varname>=0.8.1, <1.0.0",
    "python-dotenv>=0.19.2, <=1.0.0",
    "pynrrd>=0.4.2, <1.0.0",
    "SimpleITK>=2.1.1.2, <3.0.0.0",
    "pydicom>=2.3.0, <3.0.0",
    "stringcase>=1.2.0, <2.0.0",
    "python-magic>=0.4.25, <1.0.0",
    "trimesh>=3.11.2, <4.0.0",
    "uvicorn[standard]>=0.18.2, <1.0.0",
    "pydantic>=1.7.4, <=2.5.0",
    "anyio>=3.7.1,<=4.2.0",  # TODO: remove after upgrade fastapi version up to 0.103.1
    "fastapi>=0.79.0, <=0.109.0",
    "websockets>=10.3, <11.0",
    "jinja2>=3.0.3, <4.0.0",
    "psutil>=5.9.0, <6.0.0",
    "jsonpatch>=1.32, <2.0",
    "MarkupSafe>=2.1.1, <3.0.0",
    "arel>=0.2.0, <1.0.0",
    "tqdm>=4.62.3, <5.0.0",
    "pandas>=1.1.3, <=2.1.4",
    "async_asgi_testclient",
    "PyYAML",
    "distinctipy",
    "beautifulsoup4",
    "numerize",
    "ffmpeg-python==0.2.0",
    "python-multipart==0.0.5",
    "GitPython",
    "giturlparse",
    "rich",
    "click",
    "imutils==0.5.4",
    "urllib3>=1.26.15, <=2.2.0",
    "cacheout==0.14.1",
    "jsonschema>=2.6.0,<=4.20.0",
    "pyjwt>=2.1.0,<3.0.0",
]

ALT_INSTALL_REQUIRES = {
    "opencv-python>=4.5.5.62, <5.0.0.0": [
        "opencv-python-headless",
        "opencv-contrib-python",
        "opencv-contrib-python-headless",
    ],
}


def check_alternative_installation(install_require, alternative_install_requires):
    """If some version version of alternative requirement installed, return alternative,
    else return main.
    """
    for alternative_install_require in alternative_install_requires:
        try:
            alternative_pkg_name = re.split(r"[ !<>=]", alternative_install_require)[0]
            get_distribution(alternative_pkg_name)
            return str(alternative_install_require)
        except DistributionNotFound:
            continue

    return str(install_require)


def get_install_requirements(main_requires, alternative_requires):
    """Iterates over all install requires
    If an install require has an alternative option, check if this option is installed
    If that is the case, replace the install require by the alternative to not install dual package
    """
    install_requires = []
    for main_require in main_requires:
        if main_require in alternative_requires:
            main_require = check_alternative_installation(
                main_require, alternative_requires.get(main_require)
            )
        install_requires.append(main_require)

    return install_requires


# Dependencies do not include PyTorch, so
# supervisely_lib.nn.hosted.pytorch will not work out of the box.
# If you need to invoke that part of the code, it is very likely you
# already have PyTorch installed.
setup(
    name="supervisely",
    maintainer="Max Kolomeychenko",
    version=version,
    description="Supervisely Python SDK.",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    author="Supervisely",
    author_email="support@supervise.ly",
    url="https://github.com/supervisely/supervisely",
    packages=find_packages(
        include=["supervisely_lib", "supervisely_lib.*", "supervisely", "supervisely.*"]
    ),
    package_data={
        "": ["*.html", "*.css", "*.js", "*.md"],
        "supervisely": [
            "video/*.sh",
            "app/development/*.sh",
            "imaging/colors.json.gz",
            "nn/tracker/bot_sort/configs/MOT17/*.yml",
            "nn/tracker/bot_sort/configs/MOT20/*.yml",
        ],
    },
    entry_points={
        "console_scripts": [
            "sly-release=supervisely.release.run:cli_run",
            "supervisely=supervisely.cli.cli:cli",
        ]
    },
    python_requires=">=3.8",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    install_requires=get_install_requirements(INSTALL_REQUIRES, ALT_INSTALL_REQUIRES),
    extras_require={
        "extras": [
            "docker>=5.0.3, <6.0.0",
            "imagecorruptions>=1.1.2, <2.0.0",
            "scikit-image>=0.17.1, <1.0.0",
            "matplotlib>=3.3.2, <4.0.0",
            "pascal-voc-writer>=0.1.4, <1.0.0",
            "scipy>=1.5.2, <2.0.0",
            "pandas>=1.1.3, <1.4.0",
            "ruamel.yaml==0.17.21",
        ],
        "apps": [
            "uvicorn[standard]>=0.18.2, <1.0.0",
            "fastapi>=0.79.0, <1.0.0",
            "websockets>=10.3, <11.0",
            "jinja2>=3.0.3, <4.0.0",
            "psutil>=5.9.0, <6.0.0",
            "jsonpatch>=1.32, <2.0",
            "MarkupSafe>=2.1.1, <3.0.0",
            "arel>=0.2.0, <1.0.0",
            "tqdm>=4.62.3, <5.0.0",
            "pandas>=1.1.3, <1.4.0",
        ],
        "docs": [
            "sphinx==4.4.0",
            "jinja2==3.0.3",
            "sphinx-immaterial==0.4.0",
            "sphinx-copybutton==0.4.0",
            "sphinx-autodoc-typehints==1.15.3",
            "sphinxcontrib-details-directive==0.1.0",
            "myst-parser==0.18.0",
        ],
        "sdk-no-usages": [
            "grpcio>=1.34.1, <2.0.0",
            "plotly>=4.11.0, <6.0.0",
            "psutil>=5.4.5, <6.0.0",
        ],
        # legacy dependencies
        "plugins": [
            "jsonschema>=2.6.0,<3.0.0",
        ],
        "sdk-nn-plugins": [
            "flask-restful>=0.3.7, <1.0.0",
            "Werkzeug>=1.0.1, <3.0.0",
        ],
        "aug": [
            "imgaug>=0.4.0, <1.0.0",
            "imagecorruptions>=1.1.2, <2.0.0",
        ],
        "tracking": [
            "yacs",
            "matplotlib>=3.3.2, <4.0.0",
            "scipy>=1.5.2, <2.0.0",
            "lap",
            "cython_bbox",
            "termcolor",
            "scikit-learn",
            "faiss-gpu",
            "tabulate",
            "tensorboard",
        ],
    },
)



import os
from typing import List, Tuple

import numpy as np

from supervisely import (
    Annotation,
    Label,
    ObjClass,
    ObjClassCollection,
    ProjectMeta,
    generate_free_name,
    logger,
)
from supervisely.geometry.bitmap import Bitmap
from supervisely.geometry.rectangle import Rectangle
from supervisely.imaging.image import read

MASKS_EXTENSION = ".png"

default_classes_colors = {
    "neutral": (224, 224, 192),
    "aeroplane": (128, 0, 0),
    "bicycle": (0, 128, 0),
    "bird": (128, 128, 0),
    "boat": (0, 0, 128),
    "bottle": (128, 0, 128),
    "bus": (0, 128, 128),
    "car": (128, 128, 128),
    "cat": (64, 0, 0),
    "chair": (192, 0, 0),
    "cow": (64, 128, 0),
    "diningtable": (192, 128, 0),
    "dog": (64, 0, 128),
    "horse": (192, 0, 128),
    "motorbike": (64, 128, 128),
    "person": (192, 128, 128),
    "pottedplant": (0, 64, 0),
    "sheep": (128, 64, 0),
    "sofa": (0, 192, 0),
    "train": (128, 192, 0),
    "tvmonitor": (0, 64, 128),
}


# returns mapping: (r, g, b) color -> some (row, col) for each unique color except black
def get_col2coord(img: np.ndarray) -> dict:
    img = img.astype(np.int32)
    h, w = img.shape[:2]
    colhash = img[:, :, 0] * 256 * 256 + img[:, :, 1] * 256 + img[:, :, 2]
    unq, unq_inv, unq_cnt = np.unique(colhash, return_inverse=True, return_counts=True)
    indxs = np.split(np.argsort(unq_inv), np.cumsum(unq_cnt[:-1]))
    col2indx = {unq[i]: indxs[i][0] for i in range(len(unq))}
    return {
        (col // (256**2), (col // 256) % 256, col % 256): (indx // w, indx % w)
        for col, indx in col2indx.items()
        if col != 0
    }


def read_colors(colors_file: str) -> Tuple[ObjClassCollection, dict]:
    if os.path.isfile(colors_file):
        logger.info("Will try to read segmentation colors from provided file.")
        in_lines = filter(None, map(str.strip, open(colors_file, "r").readlines()))
        in_splitted = (x.split() for x in in_lines)
        # Format: {name: (R, G, B)}, values [0; 255]
        cls2col = {}
        for x in in_splitted:
            if len(x) != 4:
                raise ValueError("Invalid format of colors file.")
            cls2col[x[0]] = (int(x[1]), int(x[2]), int(x[3]))
    else:
        logger.info("Will use default PascalVOC color mapping.")
        cls2col = default_classes_colors

    obj_classes_list = [
        ObjClass(name=class_name, geometry_type=Bitmap, color=color)
        for class_name, color in cls2col.items()
    ]

    logger.info(
        f"Determined {len(cls2col)} class(es).",
        extra={"classes": list(cls2col.keys())},
    )

    obj_classes = ObjClassCollection(obj_classes_list)
    color2class_name = {v: k for k, v in cls2col.items()}
    return obj_classes, color2class_name


def get_ann(
    item,
    color2class_name: dict,
    meta: ProjectMeta,
    bbox_classes_map: dict,
    renamed_classes=None,
) -> Annotation:
    segm_path, inst_path = item.segm_path, item.inst_path
    height, width = item.shape

    ann = Annotation(img_size=(height, width))
    if segm_path is None and inst_path is None:
        return ann

    segmentation_img = read(segm_path)

    if inst_path is not None:
        instance_img = read(inst_path)
        colored_img = instance_img
        instance_img16 = instance_img.astype(np.uint16)
        col2coord = get_col2coord(instance_img16)
        curr_col2cls = []
        for col, coord in col2coord.items():
            cls_name = color2class_name.get(tuple(segmentation_img[coord]))
            if cls_name is not None:
                if renamed_classes is not None and cls_name in renamed_classes:
                    cls_name = renamed_classes[cls_name]
            curr_col2cls.append((col, cls_name))
        curr_col2cls = {
            k: v for k, v in curr_col2cls if v is not None
        }  # _instance_ color -> class name
    else:
        colored_img = segmentation_img
        segmentation_img = segmentation_img.astype(np.uint16)
        colors = list(get_col2coord(segmentation_img).keys())
        curr_col2cls = {}
        for color in colors:
            cls_name = color2class_name.get(color)
            if cls_name is not None:
                if renamed_classes is not None and cls_name in renamed_classes:
                    cls_name = renamed_classes[cls_name]
            curr_col2cls[color] = cls_name

    for color, class_name in curr_col2cls.items():
        mask = np.all(colored_img == color, axis=2)  # exact match (3-channel img & rgb color)
        bitmap = Bitmap(data=mask)
        obj_class = ObjClass(name=class_name, geometry_type=Bitmap)

        ann = ann.add_label(Label(bitmap, obj_class))
        #  clear used pixels in mask to check missing colors, see below
        colored_img[mask] = (0, 0, 0)

    if np.sum(colored_img) > 0:
        logger.warn(
            f"Not all objects or classes are captured from source segmentation: {item.name}"
        )

    if item.ann_data is not None:
        bbox_labels = xml_to_sly_labels(item.ann_data, meta, bbox_classes_map, renamed_classes)
        ann = ann.add_labels(bbox_labels)

    return ann


def xml_to_sly_labels(
    xml_path: str,
    meta: ProjectMeta,
    bbox_classes_map: dict,
    renamed_classes=None,
) -> List[Label]:
    import xml.etree.ElementTree as ET

    labels = []
    with open(xml_path, "r") as in_file:
        tree = ET.parse(in_file)
        root = tree.getroot()

        for obj in root.iter("object"):
            cls_name = obj.find("name").text
            cls_name = bbox_classes_map.get(cls_name, cls_name)
            if renamed_classes and cls_name in renamed_classes:
                cls_name = renamed_classes[cls_name]
            obj_cls = meta.obj_classes.get(cls_name)
            if obj_cls is None:
                logger.warn(f"Class {cls_name} is not found in meta. Skipping.")
                continue
            xmlbox = obj.find("bndbox")
            bbox_coords = [float(xmlbox.find(x).text) for x in ("ymin", "xmin", "ymax", "xmax")]
            bbox = Rectangle(*bbox_coords)
            label = Label(bbox, obj_cls)
            labels.append(label)

    return labels


def update_meta_from_xml(
    xml_path: str,
    meta: ProjectMeta,
    existing_cls_names: set,
    bbox_classes_map: dict,
) -> ProjectMeta:
    import xml.etree.ElementTree as ET

    with open(xml_path, "r") as in_file:
        tree = ET.parse(in_file)
        root = tree.getroot()

        for obj in root.iter("object"):
            class_name = obj.find("name").text
            original_class_name = class_name
            obj_cls = meta.obj_classes.get(class_name)
            if obj_cls is None:
                obj_cls = ObjClass(name=class_name, geometry_type=Rectangle)
                meta = meta.add_obj_class(obj_cls)
                existing_cls_names.add(class_name)
                continue
            elif obj_cls.geometry_type == Rectangle:
                continue
            class_name = class_name + "_bbox"
            obj_cls = meta.obj_classes.get(class_name)
            if obj_cls is None:
                obj_cls = ObjClass(name=class_name, geometry_type=Rectangle)
                meta = meta.add_obj_class(obj_cls)
                existing_cls_names.add(class_name)
            elif obj_cls.geometry_type == Rectangle:
                pass
            else:
                class_name = generate_free_name(
                    existing_cls_names, class_name, extend_used_names=True
                )
                obj_cls = ObjClass(name=class_name, geometry_type=Rectangle)
                meta = meta.add_obj_class(obj_cls)
            bbox_classes_map[original_class_name] = class_name

    return meta
